import os, glob
from pathlib import Path  # , _PosixFlavour, _WindowsFlavour
from fastcore.foundation import patch_to, L


def Glob(x, extns=None, silent=False):
    files = glob.glob(x + "/*") if "*" not in x else glob.glob(x)
    if extns:
        if isinstance(extns, str):
            extns = extns.split(",")
        files = [f for f in files if any([f.endswith(ext) for ext in extns])]

    # if not silent: logger.opt(depth=1).log('INFO', '{} files found at {}'.format(len(files), x))
    return files


class P(Path):
    # _flavour = _PosixFlavour() if os.name == "posix" else _WindowsFlavour()

    def __new__(cls, *pathsegments):
        return super().__new__(cls, *pathsegments)

    def __getattr__(self, name):
        if name.startswith("_"):
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'"
            )

        # try:
        #     return getattr(self, name)
        # except:
        #     raise AttributeError(
        #         f"'{self.__class__.__name__}' object has no attribute '{name}'"
        #     )

        # Fetch the stems of all files in the directory
        fs = {stem(f): f for f in self.ls()}

        if name in fs:
            # If the stem matches, return the corresponding filename
            return fs[name]
        else:
            # Raise AttributeError if the attribute is not found
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'"
            )


def stem(fpath):
    return P(fpath).stem


def stems(folder, silent=False):
    if isinstance(folder, (str, P)):
        return L([stem(str(x)) for x in Glob(folder, silent=silent)])
    if isinstance(folder, list):
        return L([stem(x) for x in folder])


P.stems = lambda self: stems(self.ls())
ls = lambda self: L(self.iterdir())
P.ls = ls
P.__repr__ = lambda self: f"» {self}"
P.__og_dir__ = P.__dir__


@patch_to(P)
def __dir__(self):
    items = []
    for f in ls(self):
        _f = stem(f)
        items.append(_f)
    return self.__og_dir__() + items


if __name__ == "__main__":
    x = P()
    print(x.resolve())
    print(x.misc)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/pdf.ipynb.

# %% auto 0
__all__ = ['PDF']

# %% ../nbs/pdf.ipynb 2
from .loader import np, subplots, show, resize, L, Image
from fastcore.basics import ifnone

# %% ../nbs/pdf.ipynb 3
class PDF:
    """Load a PDF file from `path` as a list of images
    Use `show` function to see the images
    **WIP**"""

    def __init__(self, path, dfs=None, dpi=150):
        import fitz

        self.path = path
        self.dpi = dpi
        self.doc = fitz.open(path)
        self.ims = L([self.get_image(page_no) for page_no in range(len(self))])
        self.dfs = L(dfs) if dfs is not None else L([None] * len(self))

    def get_image(self, page_no, dpi=None):
        page = self.doc.load_page(page_no)
        pix = page.get_pixmap(dpi=ifnone(dpi, self.dpi))
        mode = "RGBA" if pix.alpha else "RGB"
        img = Image.frombytes(mode, [pix.width, pix.height], pix.samples)
        return img

    def __len__(self):
        return len(self.doc)

    def __getitem__(self, ix):
        return self.ims[ix]

    def show(self, ix=None, ixs=None, **kwargs):
        if ixs is not None:
            assert isinstance(ixs, (list, L))
            subplots(L(self.ims)[ixs], **kwargs)
        if ix is not None:
            show(self.ims[ix], **kwargs)
            return

        if len(self) == 1:
            show(self.ims[0], df=self.dfs[0], **kwargs)
        else:
            subplots(self.ims, dfs=self.dfs, **kwargs)



import traceback
from contextlib import contextmanager
from functools import partial

from fastcore.basics import ifnone

from ..loader import *
from ..markup2 import AD
from ..torch_loader import *


class IOHook:
    def __init__(self, module, depth=None, kwarg_hooks=None):
        self.module = module
        self.hook_handles = []
        self.depth = ifnone(depth, 1)
        self.kwarg_hooks = ifnone(kwarg_hooks, {})
        self.register_hooks()

    def hook_fn(self, module, input, kwargs, output, name):
        Info(f"{name} ({module.__class__.__name__})")
        log_depth = 2
        if hasattr(module, "weight"):
            Info(f"Weights: {module.weight}", depth=log_depth)
        if hasattr(module, "bias"):
            Info(f"Biases: {module.bias}", depth=log_depth)
        if input:
            Info(f"Input: {input}", depth=log_depth)
        if kwargs:
            Info(f"KWargs: {kwargs}", depth=log_depth)
        Info(f"Output: {output}", depth=log_depth)
        line()

    def register_hooks_recursive(self, submodule, prefix, current_depth=1):
        if self.depth is None or current_depth <= self.depth:
            hook_handle = submodule.register_forward_hook(
                partial(self.hook_fn, name=prefix),
                with_kwargs=True,
            )
            Info(f"Registered {submodule.__class__.__name__} @ {prefix}")
            self.hook_handles.append(hook_handle)

        if self.depth is None or current_depth < self.depth:
            for name, subsubmodule in submodule.named_children():
                self.register_hooks_recursive(
                    subsubmodule,
                    prefix=prefix + "." + name,
                    current_depth=current_depth + 1,
                )

    def register_hooks(self):
        self.register_hooks_recursive(self.module, prefix="")

    def remove_hooks(self):
        for handle in self.hook_handles:
            handle.remove()


# skip = {'SiLU', 'LayerNorm'}
# keep = {'CrossAttnDownBlock2D', 'CrossAttnUpBlock2D', 'UNetMidBlock2DCrossAttn', 'UpBlock2D', 'DownBlock2D'}
def print_shapes_hook(
    module, input, kwargs, output, skip=None, keep=None, child_name=""
):
    from builtins import print

    skip = set() if skip is None else set(skip)  # Modules to exclude while printing
    keep = set() if keep is None else set(keep)  # Modules to include while printing
    try:
        name = module.__class__.__name__
        if name in skip:
            return
        if keep != {"all"} and name not in keep:
            return
        line()
        print(f"Module Name: {name}. Child Name: {child_name}")
        print(f"Input Kwargs: ")
        if "keys" in kwargs:  # kwargs literlly has the key called `key`
            kwargs["keys_"] = kwargs["keys"]
            del kwargs["keys"]
        print(AD(kwargs).summary(depth=1))
        print(f"Input Args:")
        if isinstance(input, (list, tuple, set)):
            input = {str(i + 1): v for i, v in enumerate(input)}
            print(AD(input).summary(depth=1))
        else:
            input = {"1": input}
            print(AD(input).summary(depth=1))
        print("Outputs: ")
        if isinstance(output, (list, tuple, set)):
            output = {str(i + 1): v for i, v in enumerate(output)}
            print(AD(output).summary(depth=1))
        else:
            output = {"1": output}
            print(AD(output).summary(depth=1))
        line()
    except Exception as e:
        print(f"ERROR: {e} @ {name}")


def attach_hooks(
    model,
    hook=print_shapes_hook,
    required_children=None,
    print_only=None,
    all_children=False,
    parent_name="",
):
    """Function to attach the hooks and return handles"""
    # handles = [model.register_forward_hook(hook, with_kwargs=True)]
    handles = []
    for child_name, layer in model.named_children():
        _child_name = f"{parent_name}.{child_name}" if parent_name != "" else child_name
        if (
            all_children
            or (required_children is not None and _child_name in required_children)
            or (layer.__class__.__name__ in print_only)
        ):
            _hook = partial(
                hook,
                keep={layer.__class__.__name__},
                child_name=_child_name,
            )
            handle = layer.register_forward_hook(_hook, with_kwargs=True)
            handles.append(handle)
        if len(list(layer.children())) > 0:
            handles.extend(
                attach_hooks(
                    layer,
                    hook=hook,
                    required_children=required_children,
                    print_only=print_only,
                    parent_name=_child_name,
                    all_children=all_children,
                )
            )
    return handles


def detach_hooks(handles):
    """Function to detach hooks"""
    for handle in handles:
        handle.remove()


@contextmanager
def hook_context_for(module, hook_fn, **kwargs):
    try:
        hook_fn = partial(hook_fn, **kwargs)
        handles = attach_hooks(module, hook_fn)
        yield module
        detach_hooks(handles)
    except Exception as e:
        Warn(f"Error: {e}")
    finally:
        try:
            detach_hooks(handles)
        except:
            # nothing needs to be done
            ...


@contextmanager
def print_module_ios_for(module, print_only={}, required_children=None):
    if print_only is None:
        print_only = {module.__class__.__name__}
    try:
        # _print_shapes_hook = partial(print_shapes_hook, keep=print_only)
        handles = attach_hooks(
            module,
            print_shapes_hook,
            required_children=required_children,
            print_only=print_only,
        )
        yield module
        detach_hooks(handles)
    except Exception as e:
        Warn(f"Error: {e}\n{traceback.format_exc()}")
    finally:
        try:
            detach_hooks(handles)
        except:
            # nothing needs to be done
            ...



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/bokeh_plotting.ipynb.

# %% auto 0
__all__ = ['parse_sz', 'get_bplot']

# %% ../nbs/bokeh_plotting.ipynb 2
from bokeh.io import output_notebook, show as bshow
from bokeh.plotting import figure, from_networkx
from bokeh.models import (
    Circle,
    Rect,
    WheelZoomTool,
    PanTool,
    BoxZoomTool,
    ResetTool,
    MultiLine,
    NodesAndLinkedEdges,
    EdgesAndLinkedNodes,
    HoverTool,
    TapTool,
    BoxSelectTool,
)
from bokeh.palettes import Spectral7
import numpy as np

# %% ../nbs/bokeh_plotting.ipynb 3
def parse_sz(size):
    """
    Parses the size argument and returns a tuple of width and height.

    Args:
        size (int or tuple): The size argument to be parsed.

    Returns:
        tuple: A tuple of width and height.

    Raises:
        NotImplementedError: If the size argument is not an int or a tuple of length 2.
    """
    if isinstance(size, int):
        return size, size
    elif isinstance(size, tuple):
        if len(size) == 2:
            return size
    raise NotImplementedError(f"function is not implemented for {size}")


def get_bplot(sz=500, **kwargs):
    """
    Create a Bokeh plot with specified size and tools.

    Parameters:
    - sz (int): Size of the plot in pixels.
    - **kwargs: Additional keyword arguments for customizing the plot.

    Returns:
    - plot (bokeh.plotting.Figure): Bokeh plot object.

    """
    h, w = parse_sz(sz)
    output_notebook()
    plot = figure(
        tools=[WheelZoomTool(), PanTool(), BoxZoomTool(), ResetTool()],
        plot_width=w,
        plot_height=h,
        match_aspect=kwargs.pop("match_aspect", False),
    )
    plot.add_tools(
        HoverTool(
            tooltips=[("index", "$index")]
            + [(i, f"@{i}") for i in kwargs.get("tooltips", [])]
        ),
        TapTool(),
        BoxSelectTool(),
    )

    plot.toolbar.active_scroll = plot.select_one(WheelZoomTool)
    plot.toolbar.active_drag = plot.select_one(PanTool)
    return plot



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/charts.ipynb.

# %% auto 0
__all__ = ['alt', 'Chart', 'CM', 'radar', 'confusion_matrix', 'spider', 'upsetaltair_top_level_configuration', 'UpSetAltair']

# %% ../nbs/charts.ipynb 2
import altair as alt
from altair import Chart

alt = alt
Chart = Chart
from .loader import *

init_plt()

# %% ../nbs/charts.ipynb 5
def confusion_matrix(df=None, truth=None, pred=None, mapping=None, save_to=None):
    """
    Generate a confusion matrix chart based on the given DataFrame or truth/prediction arrays.

    Parameters:
    - df (DataFrame, optional): The input DataFrame containing the truth and prediction columns. If not provided, the truth and prediction arrays must be provided separately.
    - truth (array-like, optional): The array-like object containing the true labels.
    - pred (array-like, optional): The array-like object containing the predicted labels.
    - mapping (dict, optional): A dictionary mapping the labels to their corresponding names.
    - save_to (str, optional): The file path to save the chart in HTML format.

    Returns:
    - chart (altair.Chart): The confusion matrix chart.

    Note:
    - If `df` is not provided, `truth` and `pred` must be provided separately.
    - If `save_to` is provided, the chart will be saved at the specified file path in HTML format.
    - If `mapping` is provided, the labels in the chart will be replaced with their corresponding names.

    Example usage:
    ```
    # Generate confusion matrix from DataFrame
    confusion_matrix(df=my_df, mapping=my_mapping, save_to="confusion_matrix.html")

    # Generate confusion matrix from separate truth and prediction arrays
    confusion_matrix(truth=my_truth, pred=my_pred, mapping=my_mapping)
    ```
    """
    if df is None:
        df = pd.DataFrame({"truth": truth, "pred": pred})
        truth = "truth"
        pred = "pred"
    DF = df[[truth, pred]]

    try:
        from sklearn.metrics import classification_report

        print(classification_report(DF[truth], DF[pred]))
    except:
        logger.info("Skipping Report")
    df = DF.pivot_table(index=truth, columns=pred, aggfunc=len).reset_index()
    df = df.melt(id_vars=truth, var_name=pred, value_name="value")
    sz = 450 if len(DF[truth].unique()) > 4 else 250

    base = alt.Chart(df, height=sz, width=sz)
    hm = base.mark_rect().encode(
        x=f"{pred}:O",
        y=f"{truth}:O",
        color=alt.Color("value:Q", scale=alt.Scale(scheme="lightorange"), legend=None),
    )

    tx = base.mark_text(baseline="middle").encode(
        x=f"{pred}:O",
        y=f"{truth}:O",
        text="value:Q",
        color=alt.condition(
            alt.datum.value > 0,
            alt.value("black"),
            alt.value("rgba(0, 0, 0, 0)"),  # Transparent color for value 0
        ),
    )

    if save_to is not None:
        assert str(save_to).endswith("html"), "Can only save in html format"
        (hm + tx).save(save_to, format="html")
        Info(f"Saved chart at {save_to}")
    else:
        return hm + tx


CM = confusion_matrix

# %% ../nbs/charts.ipynb 12
def spider(
    df,
    *,
    id_column=None,
    title=None,
    max_values=None,
    padding=1.25,
    global_scale=False,
    ax=None,
    sz=10,
):
    """
    Plot a spider chart based on the given dataframe.

    Parameters:
    - df: pandas DataFrame
        The input dataframe containing the data to be plotted.
    - id_column: str, optional
        The column name to be used as the identifier for each data point. If not provided, the index of the dataframe will be used.
    - title: str, optional
        The title of the spider chart.
    - max_values: dict, optional
        A dictionary specifying the maximum values for each category. If not provided, the maximum values will be calculated based on the data.
    - padding: float, optional
        The padding factor to be applied when calculating the maximum values. Default is 1.25.
    - global_scale: bool or float, optional
        If False, each category will have its own maximum value. If True, a single maximum value will be used for all categories. If a float value is provided, it will be used as the maximum value for all categories.
    - ax: matplotlib Axes, optional
        The axes on which to plot the spider chart. If not provided, a new figure and axes will be created.
    - sz: float, optional
        The size of the figure (both width and height) in inches. Default is 10.

    Returns:
    - None

    Example usage:
    spider(df, id_column='model', title='Spider Chart', max_values={'category1': 10, 'category2': 20}, padding=1.5)
    """
    import matplotlib.pyplot as plt

    if id_column is None:
        df = df.copy().reset_index(names="index")
        id_column = "index"
    df = df.sort_values(id_column, ascending=True)
    df = df[sorted(df.columns)]
    categories = df.dtypes[(df.dtypes == "float") | (df.dtypes == "int")].index.tolist()
    data = df[categories].to_dict(orient="list")
    ids = sorted(df[id_column].tolist())
    if max_values is None:
        if not global_scale:
            max_values = {key: padding * max(value) for key, value in data.items()}
        else:
            if isinstance(global_scale, bool):
                max_value = np.array(list(data.values())).max()
            elif isinstance(global_scale, (int, float)):
                max_value = global_scale
                padding = 1.0
            max_values = {key: padding * max_value for key, _ in data.items()}
    normalized_data = {
        key: np.array(value) / max_values[key] for key, value in data.items()
    }

    num_vars = len(data.keys())
    tiks = list(data.keys())
    tiks += tiks[:1]
    angles = np.linspace(0, 2 * np.pi, num_vars, endpoint=False).tolist() + [0]

    if ax is None:
        fig, ax = plt.subplots(figsize=(sz, sz), subplot_kw=dict(polar=True))
        _show = True
    else:
        _show = False
    for i, model_name in enumerate(ids):
        values = [normalized_data[key][i] for key in data.keys()]
        actual_values = [data[key][i] for key in data.keys()]
        values += values[:1]  # Close the plot for a better look
        ax.plot(angles, values, label=model_name)
        ax.fill(angles, values, alpha=0.15)
        for _x, _y, t in zip(angles, values, actual_values):
            t = f"{t:.2f}" if isinstance(t, float) else str(t)
            ax.text(_x, _y, t, size="xx-small")

    ax.fill(angles, np.ones(num_vars + 1), alpha=0.05)
    ax.set_yticklabels([])
    ax.set_xticks(angles)
    ax.set_xticklabels(tiks)
    ax.legend(loc="upper right", bbox_to_anchor=(0.1, 0.1))
    if title is not None:
        plt.suptitle(title)
    if _show:
        plt.show()


radar = spider

# %% ../nbs/charts.ipynb 14
# Top-level altair configuration
def upsetaltair_top_level_configuration(
    base, legend_orient="top-left", legend_symbol_size=30
):
    """
    Configure the top-level settings for an UpSet plot in Altair.

    Parameters:
    - base: The base chart to configure.
    - legend_orient: The orientation of the legend. Default is "top-left".
    - legend_symbol_size: The size of the legend symbols. Default is 30.

    Returns:
    - The configured chart.

    """
    return (
        base.configure_view(stroke=None)
        .configure_title(
            fontSize=18, fontWeight=400, anchor="start", subtitlePadding=10
        )
        .configure_axis(
            labelFontSize=14,
            labelFontWeight=300,
            titleFontSize=16,
            titleFontWeight=400,
            titlePadding=10,
        )
        .configure_legend(
            titleFontSize=16,
            titleFontWeight=400,
            labelFontSize=14,
            labelFontWeight=300,
            padding=20,
            orient=legend_orient,
            symbolType="circle",
            symbolSize=legend_symbol_size,
        )
        .configure_concat(spacing=0)
    )


def UpSetAltair(
    data=None,
    title="",
    subtitle="",
    sets=None,
    abbre=None,
    sort_by="frequency",
    sort_order="ascending",
    width=1200,
    height=700,
    height_ratio=0.6,
    horizontal_bar_chart_width=300,
    color_range=["#55A8DB", "#3070B5", "#30363F", "#F1AD60", "#DF6234", "#BDC6CA"],
    highlight_color="#EA4667",
    glyph_size=200,
    set_label_bg_size=1000,
    line_connection_size=2,
    horizontal_bar_size=20,
    vertical_bar_label_size=16,
    vertical_bar_padding=20,
):
    """This function generates Altair-based interactive UpSet plots.

    Parameters:
        - data (pandas.DataFrame): Tabular data containing the membership of each element (row) in
            exclusive intersecting sets (column).
        - sets (list): List of set names of interest to show in the UpSet plots.
            This list reflects the order of sets to be shown in the plots as well.
        - abbre (list): Abbreviated set names.
        - sort_by (str): "frequency" or "degree"
        - sort_order (str): "ascending" or "descending"
        - width (int): Vertical size of the UpSet plot.
        - height (int): Horizontal size of the UpSet plot.
        - height_ratio (float): Ratio of height between upper and under views, ranges from 0 to 1.
        - horizontal_bar_chart_width (int): Width of horizontal bar chart on the bottom-right.
        - color_range (list): Color to encode sets.
        - highlight_color (str): Color to encode intersecting sets upon mouse hover.
        - glyph_size (int): Size of UpSet glyph (⬤).
        - set_label_bg_size (int): Size of label background in the horizontal bar chart.
        - line_connection_size (int): width of lines in matrix view.
        - horizontal_bar_size (int): Height of bars in the horizontal bar chart.
        - vertical_bar_label_size (int): Font size of texts in the vertical bar chart on the top.
        - vertical_bar_padding (int): Gap between a pair of bars in the vertical bar charts.
    """

    if (data is None) or (sets is None):
        print("No data and/or a list of sets are provided")
        return
    if (height_ratio < 0) or (1 < height_ratio):
        print("height_ratio set to 0.5")
        height_ratio = 0.5
    if len(sets) != len(abbre):
        abbre = None
        print(
            "Dropping the `abbre` list because the lengths of `sets` and `abbre` are not identical."
        )

    """
    Data Preprocessing
    """
    data["count"] = 0
    data = data[sets + ["count"]]
    data = data.groupby(sets).count().reset_index()

    data["intersection_id"] = data.index
    data["degree"] = data[sets].sum(axis=1)
    data = data.sort_values(
        by=["count"], ascending=True if sort_order == "ascending" else False
    )

    data = pd.melt(data, id_vars=["intersection_id", "count", "degree"])
    data = data.rename(columns={"variable": "set", "value": "is_intersect"})

    if abbre == None:
        abbre = sets

    set_to_abbre = pd.DataFrame(
        [[sets[i], abbre[i]] for i in range(len(sets))], columns=["set", "set_abbre"]
    )
    set_to_order = pd.DataFrame(
        [[sets[i], 1 + sets.index(sets[i])] for i in range(len(sets))],
        columns=["set", "set_order"],
    )

    degree_calculation = ""
    for s in sets:
        degree_calculation += f"(isDefined(datum['{s}']) ? datum['{s}'] : 0)"
        if sets[-1] != s:
            degree_calculation += "+"

    """
    Selections
    """
    legend_selection = alt.selection_multi(fields=["set"], bind="legend")
    color_selection = alt.selection_single(fields=["intersection_id"], on="mouseover")
    opacity_selection = alt.selection_single(fields=["intersection_id"])

    """
    Styles
    """
    vertical_bar_chart_height = height * height_ratio
    matrix_height = height - vertical_bar_chart_height
    matrix_width = width - horizontal_bar_chart_width

    vertical_bar_size = min(
        30,
        width / len(data["intersection_id"].unique().tolist()) - vertical_bar_padding,
    )

    main_color = "#3A3A3A"
    brush_opacity = alt.condition(~opacity_selection, alt.value(1), alt.value(0.6))
    brush_color = alt.condition(
        ~color_selection, alt.value(main_color), alt.value(highlight_color)
    )

    is_show_horizontal_bar_label_bg = len(abbre[0]) <= 2
    horizontal_bar_label_bg_color = (
        "white" if is_show_horizontal_bar_label_bg else "black"
    )

    x_sort = alt.Sort(
        field="count" if sort_by == "frequency" else "degree", order=sort_order
    )
    tooltip = [
        alt.Tooltip("max(count):Q", title="Cardinality"),
        alt.Tooltip("degree:Q", title="Degree"),
    ]

    """
    Plots
    """
    # To use native interactivity in Altair, we are using the data transformation functions
    # supported in Altair.
    base = (
        alt.Chart(data)
        .transform_filter(legend_selection)
        .transform_pivot(
            # Right before this operation, columns should be:
            # `count`, `set`, `is_intersect`, (`intersection_id`, `degree`, `set_order`, `set_abbre`)
            # where (fields with brackets) should be dropped and recalculated later.
            "set",
            op="max",
            groupby=["intersection_id", "count"],
            value="is_intersect",
        )
        .transform_aggregate(
            # count, set1, set2, ...
            count="sum(count)",
            groupby=sets,
        )
        .transform_calculate(
            # count, set1, set2, ...
            degree=degree_calculation
        )
        .transform_filter(
            # count, set1, set2, ..., degree
            alt.datum["degree"]
            != 0
        )
        .transform_window(
            # count, set1, set2, ..., degree
            intersection_id="row_number()",
            frame=[None, None],
        )
        .transform_fold(
            # count, set1, set2, ..., degree, intersection_id
            sets,
            as_=["set", "is_intersect"],
        )
        .transform_lookup(
            # count, set, is_intersect, degree, intersection_id
            lookup="set",
            from_=alt.LookupData(set_to_abbre, "set", ["set_abbre"]),
        )
        .transform_lookup(
            # count, set, is_intersect, degree, intersection_id, set_abbre
            lookup="set",
            from_=alt.LookupData(set_to_order, "set", ["set_order"]),
        )
        .transform_filter(
            # Make sure to remove the filtered sets.
            legend_selection
        )
        .transform_window(
            # count, set, is_intersect, degree, intersection_id, set_abbre
            set_order="distinct(set)",
            frame=[None, 0],
            sort=[{"field": "set_order"}],
        )
    )
    # Now, we have data in the following format:
    # count, set, is_intersect, degree, intersection_id, set_abbre

    # Cardinality by intersecting sets (vertical bar chart)
    vertical_bar = (
        base.mark_bar(color=main_color, size=vertical_bar_size)
        .encode(
            x=alt.X(
                "intersection_id:N",
                axis=alt.Axis(grid=False, labels=False, ticks=False, domain=True),
                sort=x_sort,
                title=None,
            ),
            y=alt.Y(
                "max(count):Q",
                axis=alt.Axis(grid=False, tickCount=3, orient="right"),
                title="Intersection Size",
            ),
            color=brush_color,
            tooltip=tooltip,
        )
        .properties(width=matrix_width, height=vertical_bar_chart_height)
    )

    vertical_bar_text = vertical_bar.mark_text(
        color=main_color, dy=-10, size=vertical_bar_label_size
    ).encode(text=alt.Text("count:Q", format=".0f"))

    vertical_bar_chart = (vertical_bar + vertical_bar_text).add_selection(
        color_selection
    )

    # UpSet glyph view (matrix view)
    circle_bg = (
        vertical_bar.mark_circle(size=glyph_size, opacity=1)
        .encode(
            x=alt.X(
                "intersection_id:N",
                axis=alt.Axis(grid=False, labels=False, ticks=False, domain=False),
                sort=x_sort,
                title=None,
            ),
            y=alt.Y(
                "set_order:N",
                axis=alt.Axis(grid=False, labels=False, ticks=False, domain=False),
                title=None,
            ),
            color=alt.value("#E6E6E6"),
        )
        .properties(height=matrix_height)
    )

    rect_bg = (
        circle_bg.mark_rect()
        .transform_filter(alt.datum["set_order"] % 2 == 1)
        .encode(color=alt.value("#F7F7F7"))
    )

    circle = circle_bg.transform_filter(alt.datum["is_intersect"] == 1).encode(
        color=brush_color
    )

    line_connection = (
        vertical_bar.mark_bar(size=line_connection_size, color=main_color)
        .transform_filter(alt.datum["is_intersect"] == 1)
        .encode(y=alt.Y("min(set_order):N"), y2=alt.Y2("max(set_order):N"))
    )

    matrix_view = (
        circle + rect_bg + circle_bg + line_connection + circle
    ).add_selection(
        # Duplicate `circle` is to properly show tooltips.
        color_selection
    )

    # Cardinality by sets (horizontal bar chart)
    horizontal_bar_label_bg = base.mark_circle(size=set_label_bg_size).encode(
        y=alt.Y(
            "set_order:N",
            axis=alt.Axis(grid=False, labels=False, ticks=False, domain=False),
            title=None,
        ),
        color=alt.Color(
            "set:N", scale=alt.Scale(domain=sets, range=color_range), title=None
        ),
        opacity=alt.value(1),
    )
    horizontal_bar_label = horizontal_bar_label_bg.mark_text(
        align=("center" if is_show_horizontal_bar_label_bg else "center")
    ).encode(
        text=alt.Text("set_abbre:N"), color=alt.value(horizontal_bar_label_bg_color)
    )
    horizontal_bar_axis = (
        (horizontal_bar_label_bg + horizontal_bar_label)
        if is_show_horizontal_bar_label_bg
        else horizontal_bar_label
    )

    horizontal_bar = (
        horizontal_bar_label_bg.mark_bar(size=horizontal_bar_size)
        .transform_filter(alt.datum["is_intersect"] == 1)
        .encode(
            x=alt.X(
                "sum(count):Q", axis=alt.Axis(grid=False, tickCount=3), title="Set Size"
            )
        )
        .properties(width=horizontal_bar_chart_width)
    )

    # Concat Plots
    upsetaltair = alt.vconcat(
        vertical_bar_chart,
        alt.hconcat(
            matrix_view,
            horizontal_bar_axis,
            horizontal_bar,  # horizontal bar chart
            spacing=5,
        ).resolve_scale(y="shared"),
        spacing=20,
    ).add_selection(legend_selection)

    # Apply top-level configuration
    upsetaltair = upsetaltair_top_level_configuration(
        upsetaltair, legend_orient="top", legend_symbol_size=set_label_bg_size / 2.0
    ).properties(
        title={
            "text": title,
            "subtitle": subtitle,
            "fontSize": 20,
            "fontWeight": 500,
            "subtitleColor": main_color,
            "subtitleFontSize": 14,
        }
    )

    return upsetaltair



@profile
def main():
    from torch_snippets import logger
    from torch_snippets import loader
    from torch_snippets import paths
    from torch_snippets import markup
    from torch_snippets import inspector
    from torch_snippets import load_defaults
    from torch_snippets import pdf_loader
    from torch_snippets import markup2
    from torch_snippets import registry
    from torch_snippets import ipython
    from torch_snippets import decorators
    from torch_snippets import misc
    from torch_snippets import dates

    # from torch_snippets import torch_loader
    from torch_snippets import s3_loader
    from torch_snippets import zen


print(main())



from typer import Typer
from .loader import show as _show

cli = Typer()


@cli.command()
def time():
    """print the time now, in %Y-%m-%d %H:%M:%S format"""
    import time

    print(time.strftime("%Y-%m-%d %H:%M:%S"))


# Other cool utils


@cli.command()
def show(path):
    """print the object"""
    _show(path)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/markups.ipynb.

# %% auto 0
__all__ = ['AttrDict', 'Config', 'isnamedtupleinstance', 'unpack', 'hash_tensor', 'hash_pandas_dataframe', 'AttrDictDeprecated',
           'decompose', 'pretty_json', 'read_json', 'write_json', 'write_jsonl', 'read_jsonl', 'read_yaml',
           'write_yaml', 'read_xml', 'write_xml']

# %% ../nbs/markups.ipynb 3
import json, os
from json import JSONEncoder
import jsonlines
import yaml
import hashlib

from .loader import BB, L, np
from .paths import *
from .logger import *
from .markup2 import AD
import xmltodict
from typing import Union
from .thinc_parser.parser import Config

# %% ../nbs/markups.ipynb 4
def _default(self, obj):
    import numpy as np
    from datetime import datetime, date

    if isinstance(obj, P):
        return str(obj)
    if isinstance(obj, AD):
        return obj.to_dict()
    if isinstance(obj, (set, L)):
        return list(obj)
    try:
        import torch

        if isinstance(obj, torch.Tensor):
            obj = obj.cpu().detach().numpy()
    except:
        ...
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    return getattr(obj.__class__, "__json__", _default.default)(obj)


_default.default = JSONEncoder().default
JSONEncoder.default = _default


def isnamedtupleinstance(x):
    _type = type(x)
    bases = _type.__bases__
    if len(bases) != 1 or bases[0] != tuple:
        return False
    fields = getattr(_type, "_fields", None)
    if not isinstance(fields, tuple):
        return False
    return all(type(i) == str for i in fields)


def unpack(obj):
    if isinstance(obj, dict):
        return {key: unpack(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [unpack(value) for value in obj]
    elif isnamedtupleinstance(obj):
        return {key: unpack(value) for key, value in obj._asdict().items()}
    elif isinstance(obj, tuple):
        return tuple(unpack(value) for value in obj)
    else:
        return obj


def hash_tensor(tensor, n=8):
    import torch

    assert isinstance(tensor, torch.Tensor)
    try:
        tensor_str = tensor.cpu().detach().numpy().tobytes()
    except:
        ...
    hash_obj = hashlib.sha256(tensor_str)
    return "ID:#" + hash_obj.hexdigest()[:n]


def hash_pandas_dataframe(input):
    try:
        from pandas.util import hash_pandas_object

        h = hash_pandas_object(input, index=True).values
        return "ID:#" + hashlib.sha256(h).hexdigest()[:6]
    except:
        return "ID:#<uncomputable>"


class AttrDictDeprecated(object):
    """
    Utility class to interact with a dictionary as if it were an object. `AD` is an alias to this class

    FEATURES:
    0. Access and modify keys (including nested keys) as if they were object attributes, supporting tab-completion.
       Example: `self.key1.key2[0].key3`
    1. Keys and values are recursively converted to AttrDict instances.
    2. Pretty-print the dictionary using `self.pretty()`.
    3. Convert the entire structure to a regular dictionary at any time using `self.to_dict()`.
    3. Recursively remove keys using `self.drop(key)` from a JSON object.
    4. Apply a function to all values at all levels using `map`.

    GOTCHAS:
    1. All integer keys are implicitly converted to strings due to the enforced `self.key` format.
    2. You can still use `self[int]`, but this internally converts the integer to a string.

    METHODS:
    - `items()`: Return the items of the AttrDict as key-value pairs.
    - `keys()`: Return the keys of the AttrDict.
    - `values()`: Return the values of the AttrDict.
    - `update(dict)`: Update the AttrDict with key-value pairs from another dictionary.
    - `get(key, default=None)`: Get the value associated with a key, with an optional default value.
    - `__iter__()`: Allow iteration over the keys of the AttrDict.
    - `__len__()`: Return the number of keys in the AttrDict.
    - `__repr__()`: Return a string representation of the AttrDict.
    - `__dir__()`: List the keys of the AttrDict as attributes.
    - `__contains__(key)`: Check if a key exists in the AttrDict.
    - `__delitem__(key)`: Delete a key from the AttrDict.
    - `map(func)`: Apply a function to all values in the AttrDict.
    - `drop(key)`: Recursively remove a key and its values from the AttrDict.
    - `to_dict()`: Convert the AttrDict and its nested structure to a regular dictionary.
    - `pretty(print_with_logger=False, *args, **kwargs)`: Pretty-print the AttrDict as JSON.
    - `__eq__(other)`: Compare the AttrDict with another dictionary for equality.
    - `find_address(key, current_path="")`: Find and return all addresses (paths) of a given key in the AttrDict.
    - `summary(current_path='', summary_str='', depth=0, sep='\t')`: Generate a summary of the structure and values in the AttrDict.
    - `write_summary(to, **kwargs)`: Write the summary to a file or stream.
    - `fetch(addr)`: Retrieve a value at a specified address (path).

    PARAMETERS:
    - `data` (dict, optional): Initial data to populate the AttrDict.

    USAGE:
    - Create an AttrDict instance by providing an optional initial dictionary, and then access and manipulate its contents as if they were object attributes.

    EXAMPLE:
    ```python
    my_dict = {'name': 'John', 'age': 30, 'address': {'city': 'New York', 'zip': '10001'}}
    attr_dict = AttrDict(my_dict)
    print(attr_dict.name)  # Access values like attributes
    attr_dict.address.city = 'Los Angeles'  # Modify nested values
    ```
    """

    def __init__(self, data=None, **kwargs):
        if data is None:
            data = {}
        data = {**data, **kwargs}
        for name, value in data.items():
            setattr(self, str(name), self._wrap(value))

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def _wrap(self, value):
        if isinstance(value, (L, tuple, list, set, frozenset)):
            value = type(value)([self._wrap(v) for v in value])
            if isinstance(value, (list, L)):
                value = L(value)
            return value
        else:
            return AttrDict(value) if isinstance(value, dict) else value

    __getitem__ = lambda self, x: (
        AttrDict({_x: self[_x] for _x in x})
        if isinstance(x, (list, L))
        else getattr(self, str(x))
    )
    __setitem__ = lambda self, k, v: setattr(self, str(k), self._wrap(v))

    def update(self, dict):
        for k, v in dict.items():
            self[k] = v

    def get(self, key, default=None):
        key = str(key)
        return self[key] if key in self else default

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())

    def __repr__(self):
        return self.summary()
        return "{%s}" % str(
            ", ".join("'%s': %s" % (k, repr(v)) for (k, v) in self.__dict__.items())
        )

    def __dir__(self):
        return self.__dict__.keys()

    def __contains__(self, key):
        key = str(key)
        if "." not in key:
            return key in self.__dict__.keys()
        else:
            d = self
            for _k in key.split("."):
                try:
                    d = d[_k]
                except AttributeError:
                    return False
            return True

    def __delitem__(self, key):
        key = str(key)
        del self.__dict__[key]

    def map(self, func):
        for k in dir(self):
            v = self[k]
            if isinstance(v, AttrDict):
                v.map(func)
            elif isinstance(v, (L, tuple, list, set, frozenset)):
                v = [_v.map(func) if isinstance(_v, AttrDict) else func(_v) for _v in v]
            else:
                v = func(v)
            self[k] = v

    def drop(self, key):
        if key in self:
            del self[key]
        for k in dir(self):
            v = self[k]
            if isinstance(v, AttrDict):
                v.drop(key)
            if isinstance(v, (L, tuple, list, set, frozenset)):
                v = [_v.drop(key) for _v in v if isinstance(_v, AttrDict)]

    def to_dict(self):
        d = {}
        for k in dir(self):
            v = self[k]
            if isinstance(v, AttrDict):
                v = v.to_dict()
            if isinstance(v, (L, tuple, list, set, frozenset)):
                v = [_v.to_dict() if isinstance(_v, AttrDict) else _v for _v in v]
            d[k] = v
        return d

    def pretty(self, print_with_logger=False, *args, **kwargs):
        pretty_json(
            self.to_dict(), print_with_logger=print_with_logger, *args, **kwargs
        )

    def __eq__(self, other):
        return AttrDict(other).to_dict() == self.to_dict()

    def find_address(self, key, current_path=""):
        addresses = []
        for k in self.keys():
            if current_path:
                new_path = f"{current_path}.{k}"
            else:
                new_path = k

            if k == key:
                addresses.append(new_path)

            if isinstance(self[k], AttrDict):
                addresses.extend(self[k].find_address(key, new_path))

            elif isinstance(self[k], (L, tuple, list, set, frozenset)):
                for i, item in enumerate(self[k]):
                    if isinstance(item, AttrDict):
                        addresses.extend(item.find_address(key, f"{new_path}.{i}"))
        return addresses

    def summary(self, current_path="", depth=0, sep="  ", max_items=10):
        max_items = int(os.environ.get("AD_MAX_ITEMS", max_items))
        if max_items == -1:
            max_items = 1e3
        sep = os.environ.get("AD_SEP", sep)

        def format_path(path, key):
            return f"{path}.{key}" if path else key

        def format_item(key, item, path, depth, sep):
            import numpy as np
            import pandas as pd

            try:
                import torch
            except ModuleNotFoundError:

                class Torch:
                    Tensor = type(None)

                torch = Torch()

            if isinstance(item, (pd.DataFrame,)):
                return f"{sep * depth}{key} - {type(item).__name__} - shape {item.shape} - columns {item.columns} - {hash_pandas_dataframe(item)}\n"
            if isinstance(item, AttrDict) or hasattr(item, "keys"):
                item = AttrDict(**item)
                return f"{sep*depth}{key}\n" + item.summary(path, depth + 1, sep)
            elif isinstance(item, (list, tuple, set, frozenset, L)):
                return summarize_collection(key, item, path, depth + 1, sep)
            elif isinstance(item, (torch.Tensor, np.ndarray)):
                is_np = False
                if isinstance(item, np.ndarray):
                    is_np = True
                    item = torch.tensor(item)
                is_np = "🔦" if not is_np else "np."
                return f"{sep * depth}{key} - {is_np}{item} - {hash_tensor(item)}\n"
            elif isinstance(item, type):
                raise NotImplementedError("Implement for type object")

            else:
                is_multiline = False
                ogitem = item
                if isinstance(item, (str, P)):
                    item = str(item)
                    is_multiline = "\n" in item
                    _sep = " ...\n...\n...\n...\n... " if is_multiline else "........."
                    if len(item) > 100:
                        item = item[:35] + _sep + item[-35:]
                    if is_multiline:
                        _item = item.split("\n")
                        _item = "\n".join([f"{sep*(depth+1)}{l}" for l in _item])
                        item = f"↓\n{sep*(depth+1)}```\n{_item}\n{sep*(depth+1)}```"
                multiline = "" if not is_multiline else "Multiline "
                return f"{sep * depth}{key} - {item} (🏷️ {multiline}{type(ogitem).__name__})\n"

        def summarize_collection(key, collection, path, d, s):
            summary_str = f"{s * (d - 1)}{key}\n"
            for i, item in enumerate(collection):
                item_path = format_path(path, i)
                if i < max_items:
                    summary_str += format_item(i, item, item_path, d, s)
                else:
                    summary_str += (
                        f"{s*d}... {len(collection) - max_items} more items ...\n"
                    )
                    break
            return summary_str

        summary_str = ""
        for ix, key in enumerate(self.keys()):
            if ix >= max_items:
                summary_str += (
                    f"{sep*depth} ... {len(self.keys()) - max_items} more keys ...\n"
                )
                break
            new_path = format_path(current_path, key)
            summary_str += format_item(key, self[key], new_path, depth, sep)
        return summary_str

    def write_summary(self, to, **kwargs):
        writelines(self.summary(**kwargs).split("\n"), to, "w")

    def fetch(self, addr):
        if isinstance(addr, (list, L)):
            return L([self.fetch(_addr) for _addr in addr])

        o = self
        for p in addr.split("."):
            try:
                o = o[int(p)]
            except:
                o = o[p]
        return o


AttrDict = AD


def decompose(i):
    print(
        AD(
            {k: getattr(i, k) for k in dir(i) if not k.startswith("_")},
            type=str(type(i)),
        )
    )


def pretty_json(
    i, fpath=None, indent=4, print_with_logger=True, return_as_string=False
):
    def set_default(obj):
        if isinstance(obj, (set, BB, L)):
            return list(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, P):
            return str(obj)

    assert isinstance(i, (dict, list))
    i = unpack(i)
    dump = json.dumps(i, indent=indent, default=set_default)
    if fpath:
        makedir(parent(fpath))
        logger.info(f"Dumped a pretty file to {fpath}")
        with open(fpath, "w") as f:
            json.dump(i, f, indent=indent, default=set_default)
            return
    if print_with_logger:
        return logger.opt(depth=1).log("DEBUG", f"\n{dump}")
    else:
        if return_as_string:
            return dump
        print(dump)

# %% ../nbs/markups.ipynb 9
def read_json(fpath):
    import json

    with open(fpath, "r") as f:
        return json.load(f)


def write_json(obj, fpath, silent=False):
    if not silent:
        logger.opt(depth=1).log("DEBUG", f"Dumping json to {fpath}")
    with open(fpath, "w") as f:
        json.dump(obj, f, indent=4)
    return P(fpath)

# %% ../nbs/markups.ipynb 12
def write_jsonl(items, dest, mode="a"):
    makedir(parent(dest))
    with jsonlines.open(dest, mode) as writer:
        writer.write_all(items)
        if mode == "a":
            Info(f"Appended {len(items)} items to {dest}")
        elif mode == "w":
            Info(f"Wrote {len(items)} jsons to {dest}")


def read_jsonl(file):
    return [json.loads(line) for line in readlines(file, silent=True)]

# %% ../nbs/markups.ipynb 13
def read_yaml(file):
    with open(file, "r") as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)


def write_yaml(content, fpath):
    with open(fpath, "w") as outfile:
        yaml.dump(content, outfile, default_flow_style=False)

# %% ../nbs/markups.ipynb 14
def read_xml(file_path: Union[str, P]) -> AttrDict:
    "Read xml data as a dictionary"
    with open(str(file_path)) as xml_file:
        data = AttrDict(xmltodict.parse(xml_file.read()))
    return data


def write_xml(data: Union[AttrDict, dict], file_path: Union[str, P]):
    makedir(parent(file_path))
    "convert a dictionary to xml"
    with open(file_path, "w") as xml_file:
        data = data.to_dict() if isinstance(data, AttrDict) else data
        assert isinstance(data, dict), "Function only supports dicts for now"
        data = xmltodict.unparse(data, pretty=True)
        xml_file.write(data)

# %% ../nbs/markups.ipynb 15
Config = Config



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/misc.ipynb.

# %% auto 0
__all__ = ['Timer', 'track2', 'summarize_input', 'timeit', 'io', 'tryy']

# %% ../nbs/misc.ipynb 2
import time
from .logger import Debug, Excep, debug_mode, Info, Trace
from .markup2 import AD
from functools import wraps
from fastcore.basics import ifnone
from fastcore.foundation import L

# %% ../nbs/misc.ipynb 3
class Timer:
    def __init__(self, N, smooth=True, mode=1):
        "print elapsed time every iteration and print out remaining time"
        "assumes this timer is called exactly N times or less"
        self.tok = self.start = time.time()
        self.N = N
        self.ix = 0
        self.smooth = smooth
        self.mode = mode
        # 0 = instant-speed, i.e., time remaining is a funciton of only the last iteration
        # Useful when you know that each loop takes unequal time (increasing/decreasing speeds)
        # 1 = average, i.e., time remaining is a function of average of all iterations
        # Usefule when you know on average each loop or a group of loops take around the same time

    def __call__(self, ix=None, info=None):
        ix = self.ix if ix is None else ix
        info = "" if info is None else f"{info}\t"
        tik = time.time()
        elapsed = tik - self.start

        if self.mode == 0:
            ielapsed = tik - self.tok
            ispeed = ielapsed
            iremaining = (self.N - (ix + 1)) * ispeed

            iunit = "s/iter"
            if ispeed < 1:
                ispeed = 1 / ispeed
                iunit = "iters/s"
            iestimate = iremaining + elapsed
            _info = f"{info}{ix+1}/{self.N} ({elapsed:.2f}s - {iremaining:.2f}s remaining - {ispeed:.2f} {iunit})"

        else:
            speed = elapsed / (ix + 1)
            remaining = (self.N - (ix + 1)) * speed
            unit = "s/iter"
            if speed < 1:
                speed = 1 / speed
                unit = "iters/s"
            estimate = remaining + elapsed
            # print(f'N={self.N} e={elapsed:.2f} _rem={self.N-(ix+1)} r={remaining:.2f} s={speed:.2f}\t\t')
            _info = f"{info}{ix+1}/{self.N} ({elapsed:.2f}s - {remaining:.2f}s remaining - {speed:.2f} {unit})"

        print(
            _info + " " * 10,
            end="\r",
        )
        self.ix += 1
        self.tok = tik


def track2(iterable, *, total=None):
    try:
        total = ifnone(total, len(iterable))
    except:
        ...
    timer = Timer(total)
    for item in iterable:
        info = yield item
        timer(info=info)
        if info is not None:
            yield  # Just to ensure the send operation stops

# %% ../nbs/misc.ipynb 11
def summarize_input(args, kwargs, outputs=None):
    o = AD(args, kwargs)
    if outputs is not None:
        o.outputs = outputs
    return o.summary()


def timeit(func):
    def inner(*args, **kwargs):
        s = time.time()
        o = func(*args, **kwargs)
        Info(f"{time.time() - s:.2f} seconds to execute `{func.__name__}`")
        return o

    return inner


def io(func=None, *, level="debug"):
    logfuncs = {
        "debug": lambda i: Debug(i, depth=1),
        "info": lambda i: Info(i, depth=1),
        "trace": lambda i: Trace(i, depth=1),
    }
    try:
        logfunc = logfuncs[level.lower()]
    except KeyError:
        raise ValueError(f"level should be one of {list(logfuncs.keys())}")

    def decorator(func):
        def inner(*args, **kwargs):
            s = time.time()
            o = func(*args, **kwargs)
            info = f"""
{time.time() - s:.2f} seconds to execute `{func.__name__}`
{summarize_input(args=args, kwargs=kwargs, outputs=o)}
            """
            logfunc(info)
            return o

        return inner

    if func is None:
        return decorator
    else:
        return decorator(func)

# %% ../nbs/misc.ipynb 19
def tryy(
    func=None,
    *,
    output_to_return_on_fail=None,
    silence_errors=False,
    print_traceback=False,
    store_errors: bool = True,
):
    def decorator(f):
        if isinstance(store_errors, bool) and store_errors:
            error_store = []
        elif isinstance(store_errors, (list, L)):
            error_store = store_errors

        def inner(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                if not print_traceback:
                    tb = f"{type(e).__name__}: {str(e)}"
                else:
                    import traceback

                    tb = traceback.format_exc()
                if not silence_errors:
                    Excep(
                        f"Error for `{f.__name__}` with \n{summarize_input(args, kwargs)}\n{tb}"
                    )
                if store_errors is not None:
                    error_store.append(
                        AD(
                            func=f.__name__,
                            args=args,
                            kwargs=kwargs,
                            tb=tb,
                            err_type=type(e).__name__,
                        )
                    )
                return output_to_return_on_fail

        inner.F = f  # Store reference to the original function
        inner.error_store = error_store

        def error_summary():
            import pandas as pd

            return pd.DataFrame([e.dict() for e in error_store])

        inner.error_summary = error_summary
        return inner

    if callable(func):
        return decorator(func)
    return decorator



from .hooks import attach_hooks
from ..markup2 import AD


class ModelIO:
    def __init__(self):
        self.layers = {}

    def save(self, module, input, kwargs, output, layer_name=""):
        info = AD(input=input, kwargs=kwargs, output=output)
        self.layers[layer_name] = info



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/decorators.ipynb.

# %% auto 0
__all__ = ['format', 'warn_on_fail', 'timeit', 'io', 'check_kwargs_not_none']

# %% ../nbs/decorators.ipynb 2
from functools import wraps
from .inspector import inspect
from .logger import Info, Warn
from .loader import L
from .markup import AD
import time

# %% ../nbs/decorators.ipynb 3
def format(input):
    if isinstance(input, (list, tuple, set, L)):
        input = {str(i + 1): v for i, v in enumerate(input)}
        return AD(input)
    elif isinstance(input, (dict, AD)):
        return AD(input)
    else:
        input = {"1": input}
        return AD(input)


def warn_on_fail(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            Warn(f"Exception: {e}\n@\nArgs:\n{format(args)} Kwargs:\n{format(kwargs)}")

    return wrapper


def timeit(func):
    """
    A decorator that measures the execution time of a function.

    Args:
        func (callable): The function to be timed.

    Returns:
        callable: The wrapped function.

    Example:
        @timeit
        def my_function():
            # code to be timed
            pass

        my_function()  # prints the execution time of my_function
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        output = func(*args, **kwargs)
        end = time.time()
        Info(f"{func.__name__} took {end-start:.2f} seconds to execute")
        return output

    return wrapper


def io(func):
    """
    A decorator that inspects the inputs and outputs of a function.

    Args:
        func: The function to be decorated.

    Returns:
        The decorated function.

    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if len(args) != 0:
            inspect(args, names=["inputs:args"])
        if kwargs != {}:
            inspect(kwargs, names=["inputs:kwargs"])
        output = func(*args, **kwargs)
        inspect(output, names=["outputs"])
        return output

    return wrapper


def check_kwargs_not_none(func):
    """
    A decorator that checks if any keyword argument is None.
    Raises a ValueError if any argument is None.

    Args:
        func: The function to be decorated.

    Returns:
        The decorated function.

    Raises:
        ValueError: If any keyword argument is None.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        for key, value in kwargs.items():
            if value is None:
                raise ValueError(f"Input argument '{key}' cannot be None")
        return func(*args, **kwargs)

    return wrapper



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/adapters.ipynb.

# %% auto 0
__all__ = ['np_2_b64', 'b64_2_np', 'b64_2_file', 'bytes_2_file', 'file_2_bytes', 'csvs_2_cvat', 'cvat_2_csvs', 'df_2_yolo',
           'yolo_2_df']

# %% ../nbs/adapters.ipynb 2
import base64, cv2, numpy as np
from typing import Union
from .loader import *
from .paths import P, stems, stem, parent, makedir
from .markup import write_xml, read_xml, AttrDict

# %% ../nbs/adapters.ipynb 3
def np_2_b64(image: np.ndarray) -> str:
    """Convert a numpy image to base64 string"""
    img_str = cv2.imencode(".jpg", image)[1].tobytes()
    img_b64 = base64.b64encode(img_str).decode("utf-8")
    return img_b64


def b64_2_np(input: str) -> np.ndarray:
    """Converts a base64 encoded image to a NumPy array.

    Args:
        input (str): The base64 encoded image.

    Returns:
        np.ndarray: The NumPy array representation of the image in RGB format.
    """
    input = bytes(input, "utf-8")
    input = base64.b64decode(input)
    img_nparr = np.frombuffer(input, np.uint8)
    img_cv2 = cv2.imdecode(img_nparr, cv2.IMREAD_COLOR)
    img_rgb = cv2.cvtColor(img_cv2, cv2.COLOR_BGR2RGB)
    return img_rgb


def b64_2_file(
    input: str,  # base64 encoded string
    fpath: Union[str, P],  # Place where you want to save the file
) -> None:
    """Save a file encoded as a base64 `input` at given `fpath`"""
    with open(fpath, "wb") as pdf_file:
        input = bytes(input, "utf-8")
        input = base64.b64decode(input)
        pdf_file.write(input)
        logger.info(f"wrote pdf file to {fpath}")


def bytes_2_file(
    input: bytes,  # bytes
    fpath: Union[str, P],  # Place where you want to save the file
    silent: bool = False,
) -> None:
    """Save bytes `input` at given `fpath`"""
    with open(fpath, "wb") as file_to_write:
        file_to_write.write(input)
        if not silent:
            logger.info(f"wrote pdf file to {fpath}")


def file_2_bytes(fpath):
    with open(fpath, "rb") as f:
        output = f.read()
    return output

# %% ../nbs/adapters.ipynb 4
def _process(
    df: pd.DataFrame, label_column="readable_label", default_label="Background"
):
    """
    Process the given DataFrame and convert it into a list of records.

    Args:
        df (pd.DataFrame): The input DataFrame to be processed.
        label_column (str, optional): The column name for the label. Defaults to "readable_label".
        default_label (str, optional): The default label value. Defaults to "Background".

    Returns:
        list: A list of records, where each record is a dictionary representing a row in the DataFrame.
    """
    df["@xbr"] = df["X"]
    df["@xtl"] = df["x"]
    df["@ybr"] = df["Y"]
    df["@ytl"] = df["y"]
    df["@label"] = df[label_column] if label_column in df.columns else default_label
    df["@occluded"] = "0"
    df["@source"] = "manual"
    df["@z_order"] = "0"
    if "text" in df.columns:
        df["attribute"] = [
            [{"@name": "OCR", "#text": text if text == text else ""}]
            for text in df["text"]
        ]
    df.drop(
        [
            c
            for c in df.columns
            if c
            not in set(
                [
                    "@xbr",
                    "@xtl",
                    "@ybr",
                    "@ytl",
                    "@label",
                    "@occluded",
                    "@source",
                    "@z_order",
                    "attribute",
                ]
            )
        ],
        axis=1,
        inplace=True,
    )
    records = df.to_dict(orient="records")
    return records


def csvs_2_cvat(
    images_folder,
    csvs_folder,
    xml_output_file,
    items=None,
    parquet=False,
    relative_df=True,
    default_label="Background",
    extension="jpg",
):
    """
    Convert CSV annotations to CVAT XML format.

    Args:
        images_folder (str): Path to the folder containing the images.
        csvs_folder (str): Path to the folder containing the CSV annotations.
        xml_output_file (str): Path to the output XML file.
        items (list, optional): List of items to process. If None, all items will be processed. Defaults to None.
        parquet (bool, optional): Whether the annotations are stored in Parquet format. Defaults to False.
        relative_df (bool, optional): Whether the bounding box coordinates in the CSV are relative to the image size. Defaults to True.
        default_label (str, optional): Default label for the bounding boxes. Defaults to "Background".
        extension (str, optional): Image file extension. Defaults to "jpg".

    Returns:
        None
    """
    images_folder, csvs_folder = [P(_) for _ in [images_folder, csvs_folder]]
    data = AttrDict({"annotations": {"image": []}})
    if items is None:
        items = common(stems(images_folder), stems(csvs_folder))

    items = sorted(items)

    for ix, item in enumerate(track(items)):
        _ia = _image_annotation = AttrDict({})
        image = images_folder / f"{item}.{extension}"
        height, width = read(image).shape[:2]
        if parquet:
            df = pd.read_parquet(f"{csvs_folder}/{item}.parquet")
        else:
            df = pd.read_csv(f"{csvs_folder}/{item}.csv")
        if relative_df:
            df = to_absolute(df, height, width)
        _ia["@height"], _ia["@width"] = height, width
        _ia["@id"] = str(ix)
        _ia["@name"] = f"{item}.jpg"
        _ia["box"] = _process(df, default_label=default_label)
        data.annotations.image.append(_ia)
    write_xml(data, xml_output_file)


def _get_attribute_columns(column):
    """
    Get attribute columns from a given column.

    Args:
        column (pandas.Series): The input column.

    Returns:
        set: A set of attribute columns extracted from the input column.
    """

    def _get_columns_from_row(item):
        if item != item:
            return []
        if isinstance(item, dict):
            return [item["@name"]]
        else:
            return [_item["@name"] for _item in item]

    output = set(flatten(column.map(_get_columns_from_row)))
    return output


def _get_attribute_data(item, column_name):
    """
    Retrieves the attribute data for a given item and column name.

    Parameters:
    - item: The item to retrieve the attribute data from.
    - column_name: The name of the column to retrieve the attribute data for.

    Returns:
    - The attribute data for the given item and column name, or np.nan if not found.
    """
    if item != item:
        return item
    if isinstance(item, dict):
        if item["@name"] == column_name:
            return item.get("#text", np.nan)
        else:
            return np.nan
    elif isinstance(item, list):
        item = [_item for _item in item if _item["@name"] == column_name]
        if item:
            return item[0].get("#text", np.nan)
        else:
            return np.nan


def _cvat_ann_2_csv(ann):
    """
    Convert CVAT annotation to a pandas DataFrame in CSV format.

    Args:
        ann (dict): CVAT annotation dictionary.

    Returns:
        pandas.DataFrame: DataFrame containing the converted annotation data in CSV format.
    """
    if "box" not in ann:
        return pd.DataFrame()
    if isinstance(ann.box, AttrDict):
        rows = [a.to_dict() for a in [ann.box]]
    else:
        rows = [a.to_dict() for a in ann.box]
    df = pd.DataFrame(rows)
    df.rename(
        {
            "@xtl": "x",
            "@ytl": "y",
            "@xbr": "X",
            "@ybr": "Y",
            "@label": "readable_label",
        },
        inplace=1,
        axis=1,
    )
    for col in "xyXY":
        df[col] = df[col].map(lambda x: int(float(x)))
    df.drop(["@z_order", "@source", "@occluded"], axis=1, inplace=True)
    height, width = int(ann["@height"]), int(ann["@width"])
    df = to_relative(df, height, width)

    if "attribute" in df.columns:
        new_columns = _get_attribute_columns(df.attribute)
        for column in new_columns:
            df[column] = df["attribute"].map(lambda x: _get_attribute_data(x, column))
        df.drop(["attribute"], axis=1, inplace=True)
    return df


def cvat_2_csvs(xmlfile, csvs_folder):
    """
    Convert CVAT XML annotations to CSV files.

    Args:
        xmlfile (str): Path to the CVAT XML file.
        csvs_folder (str): Path to the folder where the CSV files will be saved.

    Returns:
        None
    """
    data = read_xml(xmlfile)
    items = data.annotations.image
    if not isinstance(items, list):
        items = [items]
    for item in items:
        try:
            df = _cvat_ann_2_csv(item)
            save_at = f'{csvs_folder}/{stem(item["@name"])}.csv'
            makedir(parent(save_at))
            df.to_csv(save_at, index=False)
        except Exception as e:
            Warn(f'{e} @ {item["@name"]}')

# %% ../nbs/adapters.ipynb 5
def df_2_yolo(df, h, w, class2id, class_column):
    yolo_data = []
    for index, row in df.iterrows():
        cls = row[class_column]
        if cls not in class2id:
            continue
        class_id = class2id[cls]
        x_center = (row["x"] + row["X"]) / 2 / w
        y_center = (row["y"] + row["Y"]) / 2 / h
        width = (row["X"] - row["x"]) / w
        height = (row["Y"] - row["y"]) / h
        yolo_data.append(
            f"{class_id} {x_center:.6f} {y_center:.6f} {width:.6f} {height:.6f}"
        )
    return yolo_data


def yolo_2_df(yolo, h, w, id2class, class_column):
    data = []

    for line in yolo:
        class_id, x_center, y_center, width, height = map(float, line.split())
        class_name = id2class[int(class_id)]

        x_center *= w
        y_center *= h
        width *= w
        height *= h

        x_min = int(x_center - (width / 2))
        y_min = int(y_center - (height / 2))
        x_max = int(x_center + (width / 2))
        y_max = int(y_center + (height / 2))

        data.append([x_min, y_min, x_max, y_max, class_name])

    df = pd.DataFrame(data, columns=["x", "y", "X", "Y", class_column])
    return df



# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {
    "np_2_b64": "adapters.ipynb",
    "b64_2_np": "adapters.ipynb",
    "b64_2_pdf": "adapters.ipynb",
    "parse_sz": "bokeh_plotting.ipynb",
    "get_bplot": "bokeh_plotting.ipynb",
    "BB": "bounding_boxes.ipynb",
    "df2bbs": "bounding_boxes.ipynb",
    "bbs2df": "bounding_boxes.ipynb",
    "bbfy": "bounding_boxes.ipynb",
    "jitter": "bounding_boxes.ipynb",
    "compute_eps": "bounding_boxes.ipynb",
    "enlarge_bbs": "bounding_boxes.ipynb",
    "shrink_bbs": "bounding_boxes.ipynb",
    "randint": "bounding_boxes.ipynb",
    "iou": "bounding_boxes.ipynb",
    "split_bb_to_xyXY": "bounding_boxes.ipynb",
    "combine_xyXY_to_bb": "bounding_boxes.ipynb",
    "to_relative": "bounding_boxes.ipynb",
    "to_absolute": "bounding_boxes.ipynb",
    "to": "capsule.ipynb",
    "train": "capsule.ipynb",
    "validate": "capsule.ipynb",
    "predict": "capsule.ipynb",
    "Capsule": "capsule.ipynb",
    "alt": "charts.ipynb",
    "Chart": "charts.ipynb",
    "confusion_matrix": "charts.ipynb",
    "CM": "charts.ipynb",
    "do": "imgaug_loader.ipynb",
    "bw": "imgaug_loader.ipynb",
    "rotate": "imgaug_loader.ipynb",
    "pad": "imgaug_loader.ipynb",
    "get_size": "imgaug_loader.ipynb",
    "rescale": "imgaug_loader.ipynb",
    "crop": "imgaug_loader.ipynb",
    "imgaugbbs2bbs": "imgaug_loader.ipynb",
    "bbs2imgaugbbs": "imgaug_loader.ipynb",
    "inspect": "inspector.ipynb",
    "is_in_notebook": "jupyter_notebook.ipynb",
    "ifexists": "load_defautls.ipynb",
    "ifnone": "load_defautls.ipynb",
    "get_console": "logging.ipynb",
    "console": "logging.ipynb",
    "print": "logging.ipynb",
    "render": "logging.ipynb",
    "logger": "logging.ipynb",
    "Debug": "logging.ipynb",
    "Info": "logging.ipynb",
    "Warn": "logging.ipynb",
    "Excep": "logging.ipynb",
    "reset_logger_width": "logging.ipynb",
    "isnamedtupleinstance": "markups.ipynb",
    "unpack": "markups.ipynb",
    "read_json": "markups.ipynb",
    "write_json": "markups.ipynb",
    "AttrDict": "registry.ipynb",
    "pretty_json": "markups.ipynb",
    "write_jsonl": "markups.ipynb",
    "read_jsonl": "markups.ipynb",
    "read_yaml": "markups.ipynb",
    "write_yaml": "markups.ipynb",
    "read_xml": "markups.ipynb",
    "write_xml": "markups.ipynb",
    "Config": "registry.ipynb",
    "input_to_str": "paths.ipynb",
    "output_to_path": "paths.ipynb",
    "size": "paths.ipynb",
    "extn": "paths.ipynb",
    "Glob": "paths.ipynb",
    "sample": "paths.ipynb",
    "mv": "paths.ipynb",
    "cp": "paths.ipynb",
    "rm": "paths.ipynb",
    "P": "paths.ipynb",
    "P.ls": "paths.ipynb",
    "P.__repr__": "paths.ipynb",
    "isdir": "paths.ipynb",
    "makedir": "paths.ipynb",
    "fname": "paths.ipynb",
    "fname2": "paths.ipynb",
    "stem": "paths.ipynb",
    "stems": "paths.ipynb",
    "parent": "paths.ipynb",
    "find": "paths.ipynb",
    "P.stems": "paths.ipynb",
    "zip_files": "paths.ipynb",
    "unzip_file": "paths.ipynb",
    "md5": "paths.ipynb",
    "remove_duplicates": "paths.ipynb",
    "readlines": "paths.ipynb",
    "read_lines": "paths.ipynb",
    "writelines": "paths.ipynb",
    "write_lines": "paths.ipynb",
    "rename_batch": "paths.ipynb",
    "dumpdill": "paths.ipynb",
    "loaddill": "paths.ipynb",
    "dill": "paths.ipynb",
    "PDF": "pdf.ipynb",
    "parse": "registry.ipynb",
    "parse_and_resolve": "registry.ipynb",
    "parse_string": "registry.ipynb",
    "registry": "registry.ipynb",
    "train_test_split": "sklegos.ipynb",
    "ColumnSelector": "sklegos.ipynb",
    "transform": "sklegos.ipynb",
    "GroupedEstimator": "sklegos.ipynb",
    "GroupedPredictor": "sklegos.ipynb",
    "EstimatorTransformer": "sklegos.ipynb",
    "MakeFrame": "sklegos.ipynb",
    "ImputeMissingValues": "sklegos.ipynb",
    "LambdaTransformer": "sklegos.ipynb",
    "Cat2Num": "sklegos.ipynb",
    "SplitDateColumn": "sklegos.ipynb",
}

modules = [
    "adapters.py",
    "bokeh_loader.py",
    "bb_utils.py",
    "trainer/capsule.py",
    "charts.py",
    "imgaug_loader.py",
    "inspector.py",
    "ipython.py",
    "load_defaults.py",
    "logger.py",
    "markup.py",
    "paths.py",
    "pdf_loader.py",
    "registry.py",
    "sklegos.py",
]

doc_url = "https://sizhky.github.io/torch_snippets/"

git_url = "https://github.com/sizhky/torch_snippets/tree/master/"


def custom_doc_links(name):
    return None



__version__ = "0.540"
from .logger import *
from .loader import *
from .paths import *
from .markup import *

# from .inspector import *
from .load_defaults import *
from .pdf_loader import PDF
from .markup2 import AD
from .registry import *
from .ipython import *
from .decorators import *
from .misc import *
from .dates import *
from .s3_loader import *
from .zen import *


def init_torch():
    from .torch_loader import (
        torch,
        th,
        torchvision,
        T,
        transforms,
        nn,
        np,
        F,
        Dataset,
        DataLoader,
        optim,
        Report,
        Reshape,
        Permute,
        device,
        save_torch_model_weights_from,
        load_torch_model_weights_to,
        detach,
        cat_with_padding,
    )

    try:
        import lovely_tensors

        lovely_tensors.monkey_patch()
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Unable to import `lovely-tensors`. `pip install lovely-tensors` for a prettier experience"
        )

    globals().update(locals())



__all__ = [
    "B",
    "Blank",
    "batchify",
    "C",
    "choose",
    "common",
    "crop_from_bb",
    "diff",
    "E",
    "flatten",
    "Image",
    "jitter",
    "L",
    "lzip",
    "line",
    "lines",
    "to_absolute",
    "to_relative",
    "enlarge_bbs",
    "shrink_bbs",
    "logger",
    "np",
    "now",
    "nunique",
    "os",
    "pad",
    "pd",
    "pdfilter",
    "pdb",
    "PIL",
    "print",
    "puttext",
    "randint",
    "rand",
    "re",
    "read",
    "readPIL",
    "rect",
    "resize",
    "rotate",
    "see",
    "show",
    "store_attr",
    "subplots",
    "sys",
    "toss",
    "track",
    "tqdm",
    "Tqdm",
    "trange",
    "unique",
    "uint",
    "write",
    "BB",
    "bbfy",
    "xywh2xyXY",
    "df2bbs",
    "bbs2df",
    # ---- Logging funcs ---- #
    "Info",
    "Warn",
    "Debug",
    "Excep",
    "reset_logger",
    "get_logger_level",
    "in_debug_mode",
    "debug_mode",
    # "display",
    "typedispatch",
    "defaultdict",
    "Counter",
    "dcopy",
    "patch_to",
    "split",
    "train_test_split",
    "init_plt",
    "init_cv2",
]

import os
import re
import sys
from builtins import print
from pathlib import Path

import numpy as np
import pandas as pd
import PIL
import tqdm
from fastcore.all import L, delegates, patch_to
from fastcore.dispatch import typedispatch
from PIL import Image

# from .bb_utils import *
# from .logger import *

from .bb_utils import (
    randint,
    BB,
    df2bbs,
    bbs2df,
    bbfy,
    jitter,
    enlarge_bbs,
    shrink_bbs,
    to_relative,
    to_absolute,
)
from .logger import (
    logger,
    Info,
    Warn,
    Debug,
    Excep,
    reset_logger,
    get_logger_level,
    in_debug_mode,
    debug_mode,
)

import datetime
import pdb
from typing import Tuple, Union

# Aliases
E = enumerate
pd.read_pqt = pd.read_parquet

from fastcore.foundation import coll_repr, is_array


@patch_to(L)
def _repr_pretty_(self, p, cycle):
    p.text(
        "..."
        if cycle
        else repr(self.items) if is_array(self.items) else coll_repr(self, 20)
    )


from collections import Counter, defaultdict
from copy import deepcopy as dcopy

from rich.progress import track as _track

track = lambda iterator, description="": _track(iterator, description=description)


old_line = lambda N=66: print("=" * N)


def line(string="", lw=66, upper=True, pad="\N{Box Drawings Double Horizontal}"):
    i = string.center(lw, pad)
    if upper:
        i = i.upper()
    print(i)


def lines(n=3, string="", **kwargs):
    assert n // 2 == (n - 1) // 2, "`n` should be odd"
    for _ in range(n // 2):
        line(**kwargs)
    line(string=string, **kwargs)
    for _ in range(n // 2):
        line(**kwargs)


def see(*X, N=66):
    list(map(lambda x: print("=" * N + "\n{}".format(x)), X)) + [print("=" * N)]


def flatten(lists):
    return [y for x in lists for y in x]


unique = lambda l: list(sorted(set(l)))
nunique = lambda l: len(set(l))


@typedispatch
def choose(List, n=1):
    if n == 1:
        return List[randint(len(List))]
    else:
        return L([choose(List) for _ in range(n)])


@typedispatch
def choose(i: dict, n=1):
    keys = list(i.keys())
    return choose(keys, n=n)


@typedispatch
def choose(i: set, n=1):
    i = list(i)
    return choose(i, n=n)


@typedispatch
def choose(i: pd.DataFrame, n=1):
    o = i.sample(n)
    if n == 1:
        return o.squeeze()
    return o


rand = lambda n=6: "".join(
    choose(list("1234567890qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM"), n=n)
)


randint = lambda high: np.random.randint(high)
randint = np.random.randint


def Tqdm(x, total=None, desc=None):
    total = len(x) if total is None else total
    return tqdm.tqdm(x, total=total, desc=desc)


from tqdm import trange

old_now = lambda: str(datetime.datetime.now())[:-10].replace(" ", "_")
now = lambda: f"{datetime.datetime.now():%Y%m%d-%H%M}"


def read(fname, mode=1):
    init_cv2()
    img = cv2.imread(str(fname), mode)
    if mode == 1:
        img = img[..., ::-1]  # BGR to RGB
    return img


def readPIL(fname, mode="RGB"):
    if mode.lower() == "bw":
        mode = "L"
    return Image.open(str(fname)).convert(mode.upper())


def crop_from_bb(im, bb, padding=None):
    if isinstance(bb, list):
        return [crop_from_bb(im, _bb, padding=padding) for _bb in bb]
    x, y, X, Y = bb
    px, py, pX, pY = padding
    if max(x, y, X, Y) < 1.5:
        h, w = im.shape[:2]
        x, y, X, Y = BB(bb).absolute((h, w))
    y = max(0, y - py)
    Y = min(h, Y + pY)
    x = max(0, x - px)
    X = min(w, X + pX)
    return im.copy()[y:Y, x:X]


def rect(im, bb, c=None, th=2):
    init_cv2()
    c = "g" if c is None else c
    _d = {"r": (255, 0, 0), "g": (0, 255, 0), "b": (0, 0, 255), "y": (255, 0, 255)}
    c = _d[c] if isinstance(c, str) else c
    x, y, X, Y = bb
    cv2.rectangle(im, (x, y), (X, Y), c, th)


def B(im, th=180):
    "Binarize Image"
    return 255 * (im > th).astype(np.uint8)


def C(im):
    "make bw into 3 channels"
    if im.shape == 3:
        return im
    else:
        return np.repeat(im[..., None], 3, 2)


def common(*items, silent=True):
    from .logger import logger

    """Wrapper around set intersection"""
    x = set(items[0])
    for item in items[1:]:
        x = set(item).intersection(x)
    lens = [str(len(i)) for i in items]
    if not silent:
        logger.opt(depth=1).log(
            "INFO",
            f"{len(x)} items found common from containers of {', '.join(lens)} items respectively",
        )
    return set(sorted(x))


def diff(a, b, rev=False, silent=False):
    from .logger import logger

    if not rev:
        o = set(sorted(set(a) - set(b)))
    else:
        o = set(sorted(set(b) - set(a)))
    if not silent:
        logger.opt(depth=1).log("INFO", f"{len(o)} items found to differ")
    return o


def puttext(im, string, org, scale=1, color=(255, 0, 0), thickness=2):
    init_cv2()
    x, y = org
    org = x, int(y + 30 * scale)
    cv2.putText(im, str(string), org, cv2.FONT_HERSHEY_COMPLEX, scale, color, thickness)


def rotate(im, angle, pad=None, return_type=np.ndarray, bbs=None):
    pad = np.median(np.array(im)) if pad is None else pad
    pad = int(pad)
    if isinstance(im, np.ndarray):
        im = Image.fromarray(im)
    im = im.rotate(angle, expand=1, fillcolor=(pad, pad, pad))
    return np.array(im)


def _jitter(i):
    return i + np.random.randint(4)


def is_in_notebook():
    try:
        import importlib

        # Test adapted from tqdm.autonotebook: https://github.com/tqdm/tqdm/blob/master/tqdm/autonotebook.py
        get_ipython = sys.modules["IPython"].get_ipython
        if "IPKernelApp" not in get_ipython().config:
            raise ImportError("console")

        return importlib.util.find_spec("IPython") is not None
    except (AttributeError, ImportError, KeyError):
        return False


def init_plt():
    import matplotlib.patheffects as path_effects
    import matplotlib.pyplot as plt

    plt.rcParams["axes.edgecolor"] = "black"
    globals().update(locals())


def init_cv2():
    import cv2

    globals().update(locals())


def show(
    img=None,
    ax=None,
    title=None,
    sz=None,
    bbs=None,
    confs=None,
    texts=None,
    bb_colors=None,
    cmap="gray",
    grid: bool = False,
    save_path: str = None,
    text_sz: int = None,
    df: pd.DataFrame = None,
    pts=None,
    conns=None,
    interactive: bool = False,
    jitter: int = None,
    frame_count: int = 1,
    **kwargs,
):
    "show an image"
    from IPython.display import display, display_html

    init_plt()

    try:
        if isinstance(img, (str, Path)):
            img = read(str(img), 1)
        try:
            import torch

            if isinstance(img, torch.Tensor):
                img = img.cpu().detach().numpy().copy()
        except ModuleNotFoundError:
            pass
        if isinstance(img, PIL.Image.Image):
            img = np.array(img)
    except Exception as e:
        Warn(e)

    if title is None:
        import inspect as I

        frame = I.currentframe()
        for _ in range(frame_count):
            frame = frame.f_back
        for var_name, var_val in frame.f_locals.items():
            if var_val is img and var_name != "_" and not var_name.strip("_").isdigit():
                title = var_name

    if isinstance(img, pd.DataFrame):
        df = img
        max_rows = kwargs.pop("max_rows", 30)
        max_rows = 10000 if max_rows == -1 else max_rows
        if is_in_notebook():
            html_str = ""
            html_str += '<th style="text-align:center"><td style="vertical-align:top">'
            if title is not None:
                html_str += f'<h3 style="text-align: center;">{title}</h3>'
            html_str += (
                df.to_html(max_rows=max_rows)
                .replace("table", 'table style="display:inline"')
                .replace(' style="display:inline"', "")
            )
            html_str += "</td></th>"
            display_html(html_str, raw=True)
        else:
            o = df.to_markdown()
            print(o)
        return
    if not isinstance(img, np.ndarray):
        display(img)
        return

    if len(img.shape) == 3 and len(img) == 3:
        # this is likely a torch tensor
        img = img.transpose(1, 2, 0)
    img = np.copy(img)
    if img.max() == 255:
        img = img.astype(np.uint8)
    h, w = img.shape[:2]

    if interactive:
        from .interactive_show import ishow

        ishow(img, df=df)
        return

    if sz is None:
        if w < 50:
            sz = 1
        elif w < 150:
            sz = 2
        elif w < 300:
            sz = 5
        elif w < 600:
            sz = 10
        else:
            sz = 20
    if isinstance(sz, int):
        sz = (sz, sz)
    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize", sz))
        _show = True
    else:
        _show = False

    if df is not None:
        if isinstance(df, (str, Path)):
            df = str(df)
            df = pd.read_csv(df) if df.endswith("csv") else pd.read_parquet(df)

        text_col = kwargs.pop("text_col", "info" if "info" in df.columns else "text")
        if text_col == "ixs":
            texts = df.index.tolist()
        elif text_col is not None and text_col in df.columns:
            texts = df[text_col]
        if "color" in df.columns:
            bb_colors = df["color"].tolist()
        bbs = df2bbs(df)  # assumes df has 'x,y,X,Y' columns or a single 'bb' column
    kwargs.pop("text_col") if "text_col" in kwargs else ...
    if isinstance(texts, pd.core.series.Series):
        texts = texts.tolist()
    if confs:
        colors = [[255, 0, 0], [223, 111, 0], [191, 191, 0], [79, 159, 0], [0, 128, 0]]
        bb_colors = [colors[int(cnf * 5) - 1] for cnf in confs]
    if isinstance(bbs, np.ndarray):
        bbs = bbs.astype(np.uint16).tolist()
    if bbs is not None:
        if "th" in kwargs:
            th = kwargs.get("th")
            kwargs.pop("th")
        else:
            if w < 800:
                th = 2
            elif w < 1600:
                th = 3
            else:
                th = 4
        if hasattr(bbs, "shape"):
            if isinstance(bbs, torch.Tensor):
                bbs = bbs.cpu().detach().numpy()
            bbs = bbs.astype(np.uint32).tolist()
        if len(bbs) > 0:
            _x_ = np.array(bbs).max()
        else:
            raise ValueError("Trying to plot with 0 bounding boxes...")
        rel = True if _x_ < 1.5 else False
        if rel:
            bbs = [BB(bb).absolute((h, w)) for bb in bbs]
        if jitter:
            bbs = [bb.jitter(jitter) for bb in bbs]
        bb_colors = (
            [[randint(255) for _ in range(3)] for _ in range(len(bbs))]
            if bb_colors == "random"
            else bb_colors
        )
        bb_colors = [bb_colors] * len(bbs) if isinstance(bb_colors, str) else bb_colors
        bb_colors = [None] * len(bbs) if bb_colors is None else bb_colors
        img = C(img) if len(img.shape) == 2 else img
        [rect(img, tuple(bb), c=bb_colors[ix], th=th) for ix, bb in enumerate(bbs)]
    text_sz = text_sz if text_sz else (max(sz) * 3 // 5)
    if texts is not None or texts == "ixs":
        if hasattr(texts, "shape"):
            if isinstance(texts, torch.Tensor):
                texts = texts.cpu().detach().numpy()
            texts = texts.tolist()
        if texts == "ixs":
            texts = [i for i in range(len(bbs))]
        if callable(texts):
            texts = [texts(bb) for bb in bbs]
        assert len(texts) == len(bbs), "Expecting as many texts as bounding boxes"
        texts = list(map(str, texts))
        texts = ["*" if len(t.strip()) == 0 else t for t in texts]
        [
            puttext(ax, text.replace("$", "\$"), tuple(bbs[ix][:2]), size=text_sz)
            for ix, text in enumerate(texts)
        ]
    if title:
        ax.set_title(title, fontdict=kwargs.pop("fontdict", None))
    if pts:
        pts = np.array(pts)
        if pts.max() < 1.1:
            pts = (pts * np.array([[w, h]])).astype(np.uint16).tolist()
        ax.scatter(*zip(*pts), c=kwargs.pop("pts_color", "red"))
    if conns is not None:
        for start_ix, end_ix, meta in conns:
            _x, _y = bbs[start_ix].xc, bbs[start_ix].yc
            _X, _Y = bbs[end_ix].xc, bbs[end_ix].yc
            _dx, _dy = _X - _x, _Y - _y
            _xc, _yc = (_X + _x) // 2, (_Y + _y) // 2
            plt.arrow(
                _jitter(_x),
                _jitter(_y),
                _jitter(_dx),
                _jitter(_dy),
                length_includes_head=True,
                color="cyan",
                head_width=4,
                head_length=4,
                width=meta * 2,
            )
            if kwargs.get("conn_text", True):
                puttext(ax, f"{meta:.2f}", (_xc, _yc), size=text_sz)
        kwargs.pop("conn_text")
    ax.imshow(img, cmap=cmap, **kwargs)

    if grid:
        ax.grid()
    else:
        ax.set_axis_off()

    if save_path:
        fig.savefig(save_path)
        return
    if _show:
        plt.show()


def puttext(ax, string, org, size=15, color=(255, 0, 0), thickness=2):
    init_plt()
    x, y = org
    va = "top" if y < 15 else "bottom"
    text = ax.text(x, y, str(string), color="red", ha="left", va=va, size=size)
    text.set_path_effects(
        [path_effects.Stroke(linewidth=3, foreground="white"), path_effects.Normal()]
    )


def subplots(ims, nc=5, figsize=(5, 5), silent=True, **kwargs):
    init_plt()
    if len(ims) == 0:
        return
    titles = kwargs.pop("titles", [None] * len(ims))
    if isinstance(titles, str):
        if titles == "ixs":
            titles = [str(i) for i in range(len(ims))]
        else:
            titles = titles.split(",")
    if len(ims) <= 5 and nc == 5:
        nc = len(ims)
    nr = (len(ims) // nc) if len(ims) % nc == 0 else (1 + len(ims) // nc)
    if not silent:
        logger.opt(depth=1).log(
            "INFO", f"plotting {len(ims)} images in a grid of {nr}x{nc} @ {figsize}"
        )
    figsize = kwargs.pop("sz", figsize)
    figsize = (figsize, figsize) if isinstance(figsize, int) else figsize
    fig, axes = plt.subplots(nr, nc, figsize=figsize)
    return_axes = kwargs.pop("return_axes", False)
    axes = axes.flat
    fig.suptitle(kwargs.pop("suptitle", ""))
    dfs = kwargs.pop("dfs", [None] * len(ims))
    bbss = kwargs.pop("bbss", [None] * len(ims))
    if "text_col" in kwargs:
        text_cols = [kwargs.pop("text_col")] * len(ims)
    else:
        text_cols = kwargs.pop("text_cols", [None] * len(ims))
    titles = titles.split(",") if isinstance(titles, str) else titles
    for ix, (im, ax) in enumerate(zip(ims, axes)):
        show(
            im,
            ax=ax,
            title=titles[ix],
            df=dfs[ix],
            bbs=bbss[ix],
            text_col=text_cols[ix],
            **kwargs,
        )
    blank = np.eye(100) + np.eye(100)[::-1]
    for ax in axes:
        show(blank, ax=ax)
    plt.tight_layout()
    if return_axes:
        return axes
    plt.show()


uint = lambda im: (255 * im).astype(np.uint8)
Blank = lambda *sh: uint(np.ones(sh))


def pdfilter(df, column, condition, silent=True):
    from .logger import logger

    if not callable(condition):
        if isinstance(condition, list):
            condition = lambda x: x in condition
        else:
            condition = lambda x: x == condition
    _df = df[df[column].map(condition)]
    if not silent:
        logger.opt(depth=1).log("DEBUG", f"Filtering {len(_df)} items out of {len(df)}")
    return _df


def pdsort(df, column, asc=True):
    df.sort_values(column, ascending=asc)


def resize(
    im: Union[np.ndarray, PIL.Image.Image],
    sz: Union[float, Tuple[int, int], Tuple[str, Tuple[int, int]]],
):
    """Resize an image based on info from sz
    *Aspect ratio is preserved
    Examples:
        >>> im = np.random.rand(100,100)
        >>> _im = resize(im, 50)                    ; assert _im.shape == (50,50)
        >>> _im = resize(im, 0.5)                   ; assert _im.shape == (50,50)   #*
        >>> _im = resize(im, (50,200))              ; assert _im.shape == (50,200)
        >>> _im = resize(im, (0.5,2.0))             ; assert _im.shape == (50,200)
        >>> _im = resize(im, (0.5,200))             ; assert _im.shape == (50,200)

        >>> im = np.random.rand(50,100)
        >>> _im = resize(im, (-1, 200))             ; assert _im.shape == (100,200) #*
        >>> _im = resize(im, (100, -1))             ; assert _im.shape == (100,200) #*
        >>> _im = resize(im, ('at-least',(40,400))) ; assert _im.shape == (200,400) #*
        >>> _im = resize(im, ('at-least',(400,40))) ; assert _im.shape == (400,800) #*
        >>> _im = resize(im, ('at-most', (40,400))) ; assert _im.shape == (40,80)   #*
        >>> _im = resize(im, ('at-most', (400,40))) ; assert _im.shape == (20,40)   #*
    """
    if isinstance(im, PIL.Image.Image):
        im = np.array(im)
        to_pil = True
    else:
        init_cv2()
        to_pil = False
    h, w = im.shape[:2]
    if isinstance(sz, (tuple, list)) and isinstance(sz[0], str):
        signal, (H, W) = sz
        assert signal in "at-least,at-most".split(
            ","
        ), "Resize type must be one of `at-least` or `at-most`"
        if signal == "at-least":
            f = max(H / h, W / w)
        if signal == "at-most":
            f = min(H / h, W / w)
        H, W = [i * f for i in [h, w]]
    elif isinstance(sz, float):
        frac = sz
        H, W = [i * frac for i in [h, w]]
    elif isinstance(sz, int):
        H, W = sz, sz
    elif isinstance(sz, tuple):
        H, W = sz
        if H == -1:
            _, W = sz
            f = W / w
            H = f * h
        elif W == -1:
            H, _ = sz
            f = H / h
            W = f * w
        elif isinstance(H, float):
            H = H * h
        elif isinstance(W, float):
            W = W * h
    H, W = int(H), int(W)
    im = cv2.resize(im, (W, H))
    if to_pil:
        im = PIL.Image.fromarray(im)
    return im


def pad(im, sz, pad_value=255):
    h, w = im.shape[:2]
    IM = np.ones(sz) * pad_value
    IM[:h, :w] = im
    return IM


def xywh2xyXY(bbs):
    if len(bbs) == 4 and isinstance(bbs[0], int):
        x, y, w, h = bbs
        return BB(x, y, x + w, y + h)
    return [xywh2xyXY(bb) for bb in bbs]


def _store_attr(self, anno, **attrs):
    for n, v in attrs.items():
        if n in anno:
            v = anno[n](v)
        setattr(self, n, v)
        self.__stored_args__[n] = v


def store_attr(names=None, self=None, but=None, cast=False, **attrs):
    "Store params named in comma-separated `names` from calling context into attrs in `self`"
    fr = sys._getframe(1)
    args = fr.f_code.co_varnames[: fr.f_code.co_argcount]
    if self:
        args = ("self", *args)
    else:
        self = fr.f_locals[args[0]]
    if not hasattr(self, "__stored_args__"):
        self.__stored_args__ = {}
    anno = self.__class__.__init__.__annotations__ if cast else {}
    if attrs:
        return _store_attr(self, anno, **attrs)
    ns = re.split(", *", names) if names else args[1:]
    but = [] if not but else but
    _store_attr(self, anno, **{n: fr.f_locals[n] for n in ns if n not in but})


def makedir(x):
    os.makedirs(x, exist_ok=True)


def parent(fpath):
    out = "/".join(fpath.split("/")[:-1])
    if out == "":
        return "./"
    else:
        return out


def write(image, fpath):
    init_cv2()
    makedir(parent(fpath))
    cv2.imwrite(fpath, image)


def lzip(*x):
    return list(zip(*x))


@patch_to(L)
def first(self):
    if len(self) > 0:
        return self[0]
    else:
        return None


@patch_to(L)
def get(self, condition):
    sublist = self.filter(condition)
    if len(sublist) > 0:
        return sublist.first()
    else:
        return None


@patch_to(L)
def get_all(self, condition):
    sublist = self.filter(condition)
    return sublist


def batchify(items, *rest, batch_size=32):
    """
    Yield batches of `batch_size` items at a time from one or more sequences.

    Parameters:
        items (iterable): The main sequence of items to be batchified.
        *rest (iterable, optional): Additional sequences that should have the same length as `items`.
        batch_size (int, optional): The number of items per batch.

    Yields:
        tuple: A tuple containing batches of items from `items` and optionally from the additional sequences.
    """
    n = len(items)
    if len(rest) > 0:
        assert all(
            [n == len(_rest) for _rest in rest]
        ), "batchify can only work with equal length containers"
    head = 0
    while head < n:
        tail = head + batch_size
        batch = items[head:tail]
        if rest:
            rest_ = [list(_rest[head:tail]) for _rest in rest]
            # yield batch, *rest_
            yield None
        else:
            yield batch
        head = tail


def toss(frac: float, pass_=True, fail_=False):
    return [fail_, pass_][np.random.uniform() < frac]


def phasify(items, n_phases: int):
    """
    [Doc generated by AI]
    Distributes a list of items into multiple phases and returns a list of lists, each containing items from a specific phase.

    Parameters:
        items (list): A list of items to be distributed among phases.
        n_phases (int): The number of phases to divide the items into.

    Returns:
        list: A list of lists, where each sublist contains items assigned to a specific phase.

    Example:
        items = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        n_phases = 3
        result = phasify(items, n_phases)
        # Output: [[1, 4, 7], [2, 5, 8], [3, 6, 9]]
    """
    iterators = defaultdict(L)
    [iterators[ix % n_phases].append(item) for ix, item in enumerate(items)]
    return L(iterators.values())


def split(items, splits, random_state=10):
    ks, vs = lzip(*splits.items())
    if any([v == -1 for v in vs]):
        assert list(vs).count(-1) == 1, "Only atmost one `-1` is allowed"
        vs = [v if v != -1 else -sum(vs) for v in vs]
    Info(vs)
    assert sum(vs) == 1, f"Split percentages should add to 1, received sum={sum(vs)}"
    np.random.seed(random_state)
    assignments = np.random.choice(ks, size=len(items), p=vs)
    o = {k: [] for k in ks}
    for item, assigned in zip(items, assignments):
        o[assigned].append(item)
    return o


def train_test_split(*args, **kwargs):
    # This is done mainly to save time and memory during imports
    from sklearn.model_selection import train_test_split as tts

    return tts(*args, **kwargs)


@patch_to(L)
def to_json(self):
    return list(self)



from .cli import cli
from .paths import P, parent, makedir
from .load_defaults import exists


@cli.command()
def download_s3_folder(s3_location, local_location, completeness_file):
    from s3fs import S3FileSystem

    s3 = S3FileSystem()
    completeness_file = P(completeness_file)
    makedir(parent(local_location))
    if not exists(completeness_file):
        s3.download(s3_location, local_location, recursive=True)
        completeness_file.touch()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/registry.ipynb.

# %% auto 0
__all__ = ['Config', 'AttrDict', 'registry', 'tryeval', 'parse_base', 'parse', 'parse_and_resolve', 'parse_string']

# %% ../nbs/registry.ipynb 2
from .markup import Config, AttrDict, L
from .thinc_parser.parser import registry
from pathlib import Path as P
from mergedeep import merge

# %% ../nbs/registry.ipynb 3
Config = Config
AttrDict = AttrDict
registry = registry


def tryeval(i):
    try:
        return eval(i)
    except:
        return i


def parse_base(o):
    if "base" not in o["META"]:
        return Config()
    _base = o["META"]["base"]
    base = P(f"{_base}.ini") if not _base.endswith(".ini") else P(_base)
    if not base.exists():
        base = P(o["META"]["file"]).parent / base
    assert (
        base.exists()
    ), f"Missing base file {base} (Please check the `base` key under `META` heading in your config. You gave `{_base}`)"
    base = parse(base, ad_format=False, interpolate=False)
    return base


def parse(file, lazyeval=False, ad_format=True, interpolate=True):
    config = Config().from_disk(file, interpolate=False)
    if "META" not in config:
        config["META"] = {}
    config["META"]["file"] = file = str(P(file).resolve())
    config["META"]["name"] = P(file).stem
    base = parse_base(config)
    config = merge(base, config)
    if interpolate:
        config = config.interpolate()
    if ad_format:
        config = AttrDict(config)
        if not lazyeval:
            config = config.map(tryeval)
    return config


def parse_and_resolve(file):
    o = AttrDict(registry.resolve(Config().from_disk(file)))
    o.map(tryeval)
    return o


def parse_string(string):
    o = AttrDict(Config().from_str(string))
    o.map(tryeval)
    return o



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/markups.ipynb.

# %% auto 0
__all__ = [
    "AD",
    "Config",
    "isnamedtupleinstance",
    "unpack",
    "AttrDict",
    "pretty_json",
    "read_json",
    "write_json",
    "write_jsonl",
    "read_jsonl",
    "read_yaml",
    "write_yaml",
    "read_xml",
    "write_xml",
]

import hashlib
from copy import deepcopy

# %% ../nbs/markups.ipynb 2
import json
import os
from collections.abc import Mapping
from json import JSONEncoder
from typing import Union, List

import jsonlines
import xmltodict
import yaml

from .icecream import ic
from .loader import BB, L, np, pd
from .logger import *
from .paths import *
from .thinc_parser.parser import Config


# %% ../nbs/markups.ipynb 3
def _default(self, obj):
    import numpy as np
    from datetime import datetime, date

    if isinstance(obj, P):
        return str(obj)
    if isinstance(obj, AD):
        return obj.to_dict()
    if isinstance(obj, (set, L)):
        return list(obj)
    try:
        import torch

        if isinstance(obj, torch.Tensor):
            obj = obj.cpu().detach().numpy()
    except:
        ...
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    return getattr(obj.__class__, "__json__", _default.default)(obj)


_default.default = JSONEncoder().default
JSONEncoder.default = _default


def isnamedtupleinstance(x):
    _type = type(x)
    bases = _type.__bases__
    if len(bases) != 1 or bases[0] != tuple:
        return False
    fields = getattr(_type, "_fields", None)
    if not isinstance(fields, tuple):
        return False
    return all(type(i) == str for i in fields)


def unpack(obj):
    if isinstance(obj, dict):
        return {key: unpack(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [unpack(value) for value in obj]
    elif isnamedtupleinstance(obj):
        return {key: unpack(value) for key, value in obj._asdict().items()}
    elif isinstance(obj, tuple):
        return tuple(unpack(value) for value in obj)
    else:
        return obj


def hash_tensor(tensor, n=8):
    import torch

    assert isinstance(tensor, torch.Tensor)
    try:
        tensor_str = tensor.cpu().detach().numpy().tobytes()
    except:
        ...
    hash_obj = hashlib.sha256(tensor_str)
    return "ID:#" + hash_obj.hexdigest()[:n]


def hash_pandas_dataframe(input):
    try:
        from pandas.util import hash_pandas_object

        h = hash_pandas_object(input, index=True).values
        return "ID:#" + hashlib.sha256(h).hexdigest()[:6]
    except:
        return "ID:#<uncomputable>"


def is_attrdict_like(input):
    o = isinstance(input, (Mapping, AttrDict)) or all(
        [hasattr(input, k) for k in ["keys", "values", "items"]]
    )
    return o


class AttrDict(object):
    """
    Utility class to interact with a dictionary as if it were an object. `AD` is an alias to this class

    FEATURES:
    0. Access and modify keys (including nested keys) as if they were object attributes, supporting tab-completion.
       Example: `self.key1.key2[0].key3`
    1. Keys and values are recursively converted to AttrDict instances.
    2. Pretty-print the dictionary using `print`.
    3. Convert the entire structure to a regular dictionary at any time using `self.to_dict() / self.dict()`.
    3. Recursively remove keys using `self.drop(key)` from a JSON object.
    4. Apply a function to all values at all levels using `map`.

    GOTCHAS:
    1. All integer keys are implicitly converted to strings due to the enforced `self.key` format.
    2. You can still use `self[int]`, but this internally converts the integer to a string.

    METHODS:
    - `items()`: Return the items of the AttrDict as key-value pairs.
    - `keys()`: Return the keys of the AttrDict.
    - `values()`: Return the values of the AttrDict.
    - `update(dict)`: Update the AttrDict with key-value pairs from another dictionary.
    - `get(key, default=None)`: Get the value associated with a key, with an optional default value.
    - `__iter__()`: Allow iteration over the keys of the AttrDict.
    - `__len__()`: Return the number of keys in the AttrDict.
    - `__repr__()`: Return a string representation of the AttrDict.
    - `__dir__()`: List the keys of the AttrDict as attributes.
    - `__contains__(key)`: Check if a key exists in the AttrDict, use 'a.b.c' notation to directly check for a nested attribute.
    - `__delitem__(key)`: Delete a key from the AttrDict.
    - `map(func)`: Apply a function to all values in the AttrDict.
    - `drop(key)`: Recursively remove a key and its values from the AttrDict.
    - `to_dict()`: Convert the AttrDict and its nested structure to a regular dictionary.
    - `pretty(print_with_logger=False, *args, **kwargs)`: Pretty-print the AttrDict as JSON.
    - `__eq__(other)`: Compare the AttrDict with another dictionary for equality.
    - `find_address(key, current_path="")`: Find and return all addresses (paths) of a given key in the AttrDict.
    - `summary(current_path='', summary_str='', depth=0, sep='\t')`: Generate a summary of the structure and values in the AttrDict.
    - `write_summary(to, **kwargs)`: Write the summary to a file or stream.
    - `fetch(addr)`: Retrieve a value at a specified address (path).

    PARAMETERS:
    - `data` (dict, optional): Initial data to populate the AttrDict.

    USAGE:
    - Create an AttrDict instance by providing an optional initial dictionary, and then access and manipulate its contents as if they were object attributes.

    EXAMPLE:
    ```python
    my_dict = {'name': 'John', 'age': 30, 'address': {'city': 'New York', 'zip': '10001'}}
    attr_dict = AttrDict(my_dict)
    print(attr_dict.name)  # Access values like attributes
    attr_dict.address.city = 'Los Angeles'  # Modify nested values
    ```
    """

    forbidden = set(":,'\"}{.")

    def __init__(self, *args, given_input_to_ad=None, **kwargs):
        given_input_to_ad = {} if given_input_to_ad is None else given_input_to_ad
        if len(args) == 1 and isinstance(args[0], (Mapping, AttrDict)):
            given_input_to_ad = args[0]
            args = {}
        else:
            _args = dict(ic.io(*args)) if len(args) > 0 else {}
            args = {}
            for k in reversed(list(_args.keys())):
                v = _args[k]
                if any(c in self.forbidden for c in k):
                    assert isinstance(
                        v, (dict, AttrDict)
                    ), f"Input `{v}` can't be a list"
                    given_input_to_ad = {**v, **given_input_to_ad}
                else:
                    args = {**{k: v}, **args}

        assert not any(
            k in list(kwargs.keys()) for k in args.keys()
        ), "Keys in args and kwargs overlap"
        given_input_to_ad = {**args, **given_input_to_ad, **kwargs}
        for name, value in given_input_to_ad.items():
            setattr(self, str(name), self._wrap(value))

    def __call__(self, **kwargs):
        o = deepcopy(self)
        for key in kwargs:
            o[key] = kwargs[key]

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def __json__(self):
        return self.to_dict()

    def _wrap(self, value):
        if isinstance(value, (L, tuple, list, set, frozenset)):
            value = type(value)([self._wrap(v) for v in value])
            if isinstance(value, (list, L)):
                value = L(value)
            return value
        else:
            return (
                AttrDict(given_input_to_ad=value) if isinstance(value, dict) else value
            )

    __getitem__ = lambda self, x: (
        AttrDict({_x: self[_x] for _x in x})
        if isinstance(x, (list, L))
        else getattr(self, str(x))
    )
    __setitem__ = lambda self, k, v: setattr(self, str(k), self._wrap(v))

    def update(self, input):
        input = AD(input)
        for k, v in input.items():
            self[k] = v

    def get(self, key, default=None):
        return self.fetch(key) if key in self else default

    def set(self, key, value):
        if "." in key:
            _keys = key.split(".")
            leaf = _keys[-1]
        else:
            _keys = [key]
            leaf = key
        for _key in _keys:
            if _key not in self and key != leaf:
                self[_key] = AD()
            if _key == leaf:
                self[_key] = value
            self = self[_key]

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())

    def __repr__(self):
        return f"\n```↯ AttrDict ↯\n{self.summary()}\n```\n"

    def __dir__(self):
        return self.__dict__.keys()

    def __contains__(self, key):
        key = str(key)
        if "." not in key:
            return key in self.__dict__.keys()
        else:
            d = self
            for _k in key.split("."):
                try:
                    d = d[_k]
                except AttributeError:
                    return False
            return True

    def __delitem__(self, key):
        key = str(key)
        del self.__dict__[key]

    def map(self, func, _try=False):
        if _try:

            def tryfunc(func):
                def inner(x):
                    try:
                        o = func(x)
                    except:
                        o = x
                    return o

                return inner

            _func = tryfunc(func)

        else:
            _func = func

        d = deepcopy(self)
        for k in dir(d):
            v = d[k]
            if isinstance(v, AttrDict):
                v = v.map(_func)
            elif isinstance(v, (L, tuple, list, set, frozenset)):
                v = [
                    _v.map(_func) if isinstance(_v, AttrDict) else _func(_v) for _v in v
                ]
            else:
                v = _func(v)
            d[k] = v
        return d

    def trymap(self, func):
        return self.map(func, _try=True)

    def drop(self, key):
        if key in self:
            del self[key]
        for k in dir(self):
            v = self[k]
            if isinstance(v, AttrDict):
                v.drop(key)
            if isinstance(v, (L, tuple, list, set, frozenset)):
                v = [_v.drop(key) for _v in v if isinstance(_v, AttrDict)]

    def to_dict(self):
        d = {}
        for k in self.__dict__.keys():  # can't use dir here
            v = self[k]
            if isinstance(v, AttrDict):
                v = v.to_dict()
            if isinstance(v, (L, tuple, list, set, frozenset)):
                v = [_v.to_dict() if isinstance(_v, AttrDict) else _v for _v in v]
            d[k] = v
        return d

    dict = to_dict

    def pretty(self, print_with_logger=False, *args, **kwargs):
        pretty_json(
            self.to_dict(), print_with_logger=print_with_logger, *args, **kwargs
        )

    def __eq__(self, other):
        return AttrDict(given_input_to_ad=other).to_dict() == self.to_dict()

    def find_address(self, key, current_path=""):
        addresses = []
        for k in self.keys():
            if current_path:
                new_path = f"{current_path}.{k}"
            else:
                new_path = k

            if k == key:
                addresses.append(new_path)

            if isinstance(self[k], AttrDict):
                addresses.extend(self[k].find_address(key, new_path))

            elif isinstance(self[k], (L, tuple, list, set, frozenset)):
                for i, item in enumerate(self[k]):
                    if isinstance(item, AttrDict):
                        addresses.extend(item.find_address(key, f"{new_path}.{i}"))
        return addresses

    def summary(self, current_path="", depth=0, sep="  ", max_items=10):
        max_items = int(os.environ.get("AD_MAX_ITEMS", max_items))
        if os.environ.get("AD_SHOW_TAB_STOPS", False):
            sep = "⋮ "
        sep = os.environ.get("AD_SEP", sep)

        def format_path(path, key):
            return f"{path}.{key}" if path else key

        def format_item(key, item, path, depth, sep):
            import numpy as np
            import pandas as pd

            try:
                import torch
            except ModuleNotFoundError:

                class Torch:
                    Tensor = type(...)

                torch = Torch()

            if isinstance(item, (pd.DataFrame,)):
                return f"{sep * depth}{key} - {type(item).__name__} - shape {item.shape} - columns {item.columns} - {hash_pandas_dataframe(item)}\n"
            if isinstance(item, AttrDict) or hasattr(item, "keys"):
                item = AttrDict(**item)
                return f"{sep*depth}{key}\n" + item.summary(path, depth + 1, sep)
            elif isinstance(item, (list, tuple, set, frozenset, L)):
                return summarize_collection(key, item, path, depth + 1, sep)
            elif isinstance(item, (torch.Tensor, np.ndarray)):
                is_np = False
                if isinstance(item, np.ndarray):
                    is_np = True
                    try:
                        item = torch.tensor(item)
                    except:
                        item = torch.tensor(item.copy())
                is_np = "🔦" if not is_np else "np."
                return f"{sep * depth}{key} - {is_np}{item} - {hash_tensor(item)}\n"

            else:
                is_multiline = False
                ogitem = item
                if isinstance(item, (str, P)):
                    item = str(item)
                    is_multiline = "\n" in item
                    _sep = " ...\n...\n...\n...\n... " if is_multiline else "........."
                    if len(item) > 100:
                        item = item[:35] + _sep + item[-35:]
                    if is_multiline:
                        _item = item.split("\n")
                        _item = "\n".join([f"{sep*(depth+1)}{l}" for l in _item])
                        item = f"↓\n{sep*(depth+1)}```\n{_item}\n{sep*(depth+1)}```"
                multiline = "" if not is_multiline else "Multiline "
                return f"{sep * depth}{key} - {item} (🏷️ {multiline}{type(ogitem).__name__})\n"

        def summarize_collection(key, collection, path, d, s):
            if isinstance(collection, (L, list)):
                info = "[]"
            elif isinstance(collection, tuple):
                info = "()"
            else:
                info = "{}"
            summary_str = f"{s * (d - 1)}{key}{info}\n"
            for i, item in enumerate(collection):
                item_path = format_path(path, i)
                if i < max_items:
                    summary_str += format_item(i, item, item_path, d, s)
                else:
                    summary_str += (
                        f"{s*d}... {len(collection) - max_items} more items ...\n"
                    )
                    break
            return summary_str

        summary_str = ""
        for ix, key in enumerate(self.keys()):
            if ix >= max_items:
                summary_str += (
                    f"{sep*depth} ... {len(self.keys()) - max_items} more keys ...\n"
                )
                break
            new_path = format_path(current_path, key)
            summary_str += format_item(key, self[key], new_path, depth, sep)
        return summary_str

    def print_summary(self, **kwargs):
        from builtins import print

        print(self.summary(**kwargs))

    def write_summary(self, to, **kwargs):
        writelines(self.summary(sep="\t", **kwargs).split("\n"), to, mode="w")

    def fetch(self, addr):
        if isinstance(addr, (list, L)):
            return L([self.fetch(_addr) for _addr in addr])

        o = self
        for p in addr.split("."):
            try:
                o = o[int(p)]
            except:
                o = o[p]
        return o

    def merge_addersses(self, ADs):
        """Given a list of ADs, merge them into a single AD, which recursively merging the addresses
        E.g. When input is
        ```↯ AttrDict ↯
        01f3daf0-f33a-11ee-a30b-129eb4343ebc.metrics.old.0.recall - 0.5 (🏷️ float)
        ```
        ,
        ```↯ AttrDict ↯
        01f3daf0-f33a-11ee-a30b-129eb4343ebc.metrics.new.0.recall - 0.7566584022717563 (🏷️ float)
        ```
        ,
        ```↯ AttrDict ↯
        01f3daf0-f33a-11ee-a30b-129eb4343ebc.metrics.new.1.recall - 0.8080808080808081 (🏷️ float)
        ```
        ,
        ```↯ AttrDict ↯
        01f3db47-f33a-11ee-a30d-129eb4343ebc.metrics.old.0.recall - 0.5 (🏷️ float)
        ```
        Output will be
        ```↯ AttrDict ↯
        01f3daf0-f33a-11ee-a30b-129eb4343ebc
          metrics
            old[]
              0
                recall - 0.5 (🏷️ float)
            new[]
              0
                recall - 0.7566584022717563 (🏷️ float)
              1
                recall - 0.8080808080808081 (🏷️ float)
        01f3db47-f33a-11ee-a30d-129eb4343ebc
          metrics
            old[]
              0
                recall - 0.5 (🏷️ float)
        ```
        """
        o = AttrDict()
        for ad in ADs:
            for k in ad.keys():
                o.set(k, ad[k])
        return o

    def flatten(self):
        """Flatten the AD into a single level AD with keys as dot combined keys"""
        o = AttrDict()
        for k in self.keys():
            v = self[k]
            if is_attrdict_like(v):
                v = v.flatten()
                for _k in v.keys():
                    o[f"{k}.{_k}"] = v[_k]
            elif isinstance(v, (list, tuple, set, frozenset, L)):
                for i, _v in enumerate(v):
                    if is_attrdict_like(_v):
                        _v = _v.flatten()
                        for _k, __v in _v.items():
                            o[f"{k}.{i}.{_k}"] = __v
                    else:
                        o[f"{k}.{i}"] = _v
            else:
                o[k] = v
        return o

    def fetch2(self, *, key=None, addrs=None):
        """given dot notation address/addresses fetch the value while maintaining the original structure"""
        if key is not None:
            addrs = self.find_address(key)
        if isinstance(addrs, (list, L)):
            return self.merge_addersses(
                L([self.fetch2(addrs=_addr) for _addr in addrs])
            )

        o = self
        for p in addrs.split("."):
            try:
                o = o[int(p)]
            except:
                o = o[p]
        return AttrDict({addrs: o})

    def write_config(self, to):
        c = Config(self.dict())
        c.to_disk(to)
        return P(to)

    def slice(self, key):
        ks = self.find_address(key)
        vs = self.fetch(ks)
        return AttrDict(dict(zip(ks, vs)))

    def flatten_and_make_dataframe(self):
        import pandas as pd

        df = pd.DataFrame([(*k.split("."), v) for k, v in self.flatten().items()])
        return df


AD = AttrDict
AD.dict = AD.to_dict


def pretty_json(
    i, fpath=None, indent=4, print_with_logger=True, return_as_string=False
):
    def set_default(obj):
        if isinstance(obj, (set, BB, L)):
            return list(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, P):
            return str(obj)

    assert isinstance(i, (dict, list))
    i = unpack(i)
    dump = json.dumps(i, indent=indent, default=set_default)
    if fpath:
        makedir(parent(fpath))
        logger.info(f"Dumped a pretty file to {fpath}")
        with open(fpath, "w") as f:
            json.dump(i, f, indent=indent, default=set_default)
            return
    if print_with_logger:
        return logger.opt(depth=1).log("DEBUG", f"\n{dump}")
    else:
        if return_as_string:
            return dump
        print(dump)


# %% ../nbs/markups.ipynb 8
def read_json(fpath):
    import json

    with open(fpath, "r") as f:
        return json.load(f)


def write_json(obj, fpath, silent=False):
    from datetime import date, datetime

    def set_default(obj):
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        raise TypeError

    if not silent:
        logger.opt(depth=1).log("DEBUG", f"Dumping json to {fpath}")
    with open(fpath, "w") as f:
        json.dump(obj, f, indent=4, default=set_default)
    return P(fpath)


# %% ../nbs/markups.ipynb 10
def write_jsonl(items, dest, mode="a"):
    makedir(parent(dest))
    with jsonlines.open(dest, mode) as writer:
        writer.write_all(items)
        if mode == "a":
            Info(f"Appended {len(items)} items to {dest}")
        elif mode == "w":
            Info(f"Wrote {len(items)} jsons to {dest}")


def read_jsonl(file):
    return [json.loads(line) for line in readlines(file, silent=True)]


# %% ../nbs/markups.ipynb 11
def read_yaml(file):
    with open(file, "r") as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)


def write_yaml(content, fpath):
    with open(fpath, "w") as outfile:
        yaml.dump(content, outfile, default_flow_style=False)


# %% ../nbs/markups.ipynb 12
def read_xml(file_path: Union[str, P]) -> AttrDict:
    "Read xml data as a dictionary"
    with open(str(file_path)) as xml_file:
        data = AttrDict(given_input_to_ad=xmltodict.parse(xml_file.read()))
    return data


def write_xml(data: Union[AttrDict, dict], file_path: Union[str, P]):
    makedir(parent(file_path))
    "convert a dictionary to xml"
    with open(file_path, "w") as xml_file:
        data = data.to_dict() if isinstance(data, AttrDict) else data
        assert isinstance(data, dict), "Function only supports dicts for now"
        data = xmltodict.unparse(data, pretty=True)
        xml_file.write(data)


def decompose(i):
    print(
        AD(
            {k: getattr(i, k) for k in dir(i) if not k.startswith("_")},
            type=str(type(i)),
        )
    )


# %% ../nbs/markups.ipynb 13
Config = Config

if __name__ == "__main__":
    assert AD({}) == {}

    p = 1
    q = {"a": 10}
    s = AD(p, q, a=20, b=30)
    assert s == {"a": 20, "b": 30, "p": 1, "q": {"a": 10}}

    s = AD(p, **q, b=30)
    assert s == {"a": 10, "b": 30, "p": 1}

    p = {"b": 222}
    assert AD(p, {"a": 2, "b": 2}, {"bb": 2, "a": 3, "b": 3}, a=20, b=30) == {
        "a": 2,
        "b": 2,
        "bb": 2,
        "p": {"b": 222},
    }

    d1 = {"c": 1}
    d2 = {"c": 2}
    d3 = {"c": 3}
    l = [1, 2, 3]
    assert AD(l, {"c": 100}, d1, d2, d3, d1={"c": 10}, d2={"c": 20}).to_dict() == {
        "l": [1, 2, 3],
        "d1": {"c": 1},
        "d2": {"c": 2},
        "c": 100,
        "d3": {"c": 3},
    }

    a = 20
    b = 30
    assert AD(a, b).to_dict() == {"a": 20, "b": 30}

    p = 10
    q = {"a": 1}
    assert AD(p, q) == {"p": 10, "q": {"a": 1}}
    assert AD(p, {"a": 1}) == {"p": 10, "a": 1}

    x = {"a": 1, "b": 2, "c": 3}
    assert AD(x).to_dict() == {"x": {"a": 1, "b": 2, "c": 3}}
    assert AD(**x).to_dict() == {"a": 1, "b": 2, "c": 3}



import os
from pathlib import Path as P

import paramiko
from loguru import logger
from scp import SCPClient as SCP


class SCPClient:
    def __init__(
        self, hostname, port, username, password=None, private_key=None, logfile=None
    ):
        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password
        self.private_key = private_key
        self.client = None
        logfile = "/tmp/{time}-scp.log" if logfile is None else logfile
        self.logger = logger
        self.logger.add(logfile)

    def connect(self):
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        if self.private_key:
            self.client.connect(
                self.hostname,
                port=self.port,
                username=self.username,
                key_filename=self.private_key,
            )
        else:
            self.client.connect(
                self.hostname,
                port=self.port,
                username=self.username,
                password=self.password,
            )

    def close(self):
        if self.client:
            self.client.close()

    def upload(self, local_path, remote_path):
        self.connect()
        try:
            with SCP(self.client.get_transport()) as scp:
                if os.path.isdir(local_path):
                    scp.put(local_path, recursive=True, remote_path=remote_path)
                else:
                    scp.put(local_path, remote_path)
                self.logger.info(f"Uploaded {local_path} to {remote_path}")
        except Exception as e:
            self.logger.warning(f"Error uploading: {e}")
        finally:
            self.close()

    def download(self, remote_path, local_path):
        self.connect()
        try:
            with SCP(self.client.get_transport()) as scp:
                is_remote_dir = "." not in remote_path
                not_a_dir = "" if is_remote_dir else "not"
                self.logger.info(f"Assuming {remote_path} is {not_a_dir}a directory")
                if is_remote_dir:
                    scp.get(remote_path, recursive=True, local_path=local_path)
                else:
                    os.makedirs(P(local_path).parent, exist_ok=True)
                    scp.get(remote_path, local_path)
                self.logger.info(f"Downloaded {remote_path} to {local_path}")
        except Exception as e:
            self.logger.warning(f"Error downloading: {e}")
        finally:
            self.close()


# Example usage
if __name__ == "__main__":
    hostname = "10.161.141.73"
    port = 22  # Default port for SSH
    username = "jioaidev"
    password = os.environ["JIOAIDEV_PASSWORD"]
    private_key = None  # Or specify the path to your private key file

    scp_client = SCPClient(hostname, port, username, password, private_key)
    local_path = "/tmp/tmp.csv"
    remote_path = "/data/datasets/210-invoices/051--2.5k-invoices-20231027/ToCleanup/vitstr_80k/00010100002900891622023/0.csv"

    # Upload a file/folder
    # scp_client.upload(local_path, remote_path)

    # Or download a file/folder
    scp_client.download(remote_path, local_path)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/config.ipynb.

# %% auto 0
__all__ = ['DeepLearningConfig', 'GenericConfig']

# %% ../../nbs/config.ipynb 1
from ..registry import parse
from .. import store_attr, ifnone
import inspect as inspect_builtin

# %% ../../nbs/config.ipynb 2
class DeepLearningConfig:
    """
    A configuration class for deep learning models.

    This class provides methods to access and manipulate configuration settings.

    Attributes:
        input_variables (list): List of input variables defined in the class constructor.

    Methods:
        keys(): Returns the list of input variables.
        __getitem__(key): Returns the value of the specified key.
        __contains__(key): Checks if the specified key is present in the input variables.
        from_ini_file(filepath, config_root=None): Creates an instance of the class from an INI file.
        __repr__(): Returns a string representation of the class.

    Example usage:
        config = DeepLearningConfig()
        config.from_ini_file('config.ini')
        print(config.keys())
        print(config['learning_rate'])
    """

    def keys(self):
        if not hasattr(self, "input_variables"):
            self.input_variables = inspect_builtin.signature(
                self.__init__
            ).parameters.keys()
        return self.input_variables

    def __getitem__(self, key):
        return getattr(self, key)

    def __contains__(self, key):
        return key in self.input_variables

    @classmethod
    def from_ini_file(cls, filepath, *, config_root=None):
        config = parse(filepath)
        config_root = ifnone(config_root, getattr(cls, "config_root"))
        if config_root is not None:
            for _root in config_root.split("."):
                config = config[_root]
        return cls(**config)

    def __repr__(self):
        return f"{self.__class__.__name__}:\n" + str({**self})

# %% ../../nbs/config.ipynb 7
class GenericConfig(DeepLearningConfig):
    def __init__(self, **kwargs):
        super().__init__()
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.input_variables = kwargs.keys()

    @classmethod
    def from_ini_file(cls, filepath, *, config_root):
        config = parse(filepath)
        for _root in config_root.split("."):
            config = config[_root]
        # convert string type to list type
        return cls(**config)
